from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
    TRANSACTIONAL_WAREHOUSE_TYPE,
    TRANSACTIONAL_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    TEST_DATABASE_CRED_MUTATION,
)


class TransactionalOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_transactional_db(self, **kwargs) -> None:
        """
        Onboard a Transactional DB connection by validating and adding a connection.
        """
        kwargs["connectionType"] = TRANSACTIONAL_CONNECTION_TYPE
        kwargs["warehouseType"] = TRANSACTIONAL_WAREHOUSE_TYPE

        self.onboard(
            validation_query=TEST_DATABASE_CRED_MUTATION,
            validation_response=EXPECTED_GENERIC_DB_GQL_RESPONSE_FIELD,
            connection_type=TRANSACTIONAL_CONNECTION_TYPE,
            **kwargs
        )



from typing import List

import click
import click_config_file

import montecarlodata.settings as settings
from montecarlodata.common import create_mc_client
from montecarlodata.integrations.configure.bi.bi import BiService
from montecarlodata.integrations.onboarding.bi.reports import ReportsOnboardingService
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.integrations.info.status import OnboardingStatusService
from montecarlodata.integrations.onboarding.data_lake.databricks import (
    DatabricksOnboardingService,
    DEFAULT_SECRET_NAME,
    DEFAULT_SECRET_SCOPE,
)
from montecarlodata.integrations.onboarding.data_lake.events import (
    EventsOnboardingService,
)
from montecarlodata.integrations.onboarding.data_lake.glue_athena import (
    GlueAthenaOnboardingService,
)
from montecarlodata.integrations.onboarding.data_lake.hive import HiveOnboardingService
from montecarlodata.integrations.onboarding.data_lake.presto import (
    PrestoOnboardingService,
)
from montecarlodata.integrations.onboarding.data_lake.spark import (
    SparkOnboardingService,
    SPARK_HTTP_MODE_CONFIG_TYPE,
    SPARK_DATABRICKS_CONFIG_TYPE,
    SPARK_BINARY_MODE_CONFIG_TYPE,
)
from montecarlodata.integrations.onboarding.etl.airflow import AirflowOnboardingService
from montecarlodata.integrations.onboarding.etl.dbt_cloud import (
    DbtCloudOnboardingService,
)
from montecarlodata.integrations.onboarding.etl.fivetran import (
    FivetranOnboardingService,
)
from montecarlodata.integrations.onboarding.fields import (
    AIRFLOW_LOGS_CONNECTION_TYPE,
    AIRFLOW_LOGS_EVENT_TYPE,
    CONNECTION_TO_WAREHOUSE_TYPE_MAP,
    GLUE_CONNECTION_TYPE,
    GQL_TO_FRIENDLY_CONNECTION_MAP,
    HIVE_MYSQL_CONNECTION_TYPE,
    HIVE_S3_CONNECTION_TYPE,
    PRESTO_S3_CONNECTION_TYPE,
    DATABRICKS_METASTORE_CONNECTION_TYPE,
    DATABRICKS_DELTA_CONNECTION_TYPE,
    S3_METADATA_EVENT_TYPE,
    S3_QL_EVENT_TYPE,
    SECRETS_MANAGER_CREDENTIAL_MECHANISM,
    SELF_HOSTING_MECHANISMS,
    SQL_SERVER_DB_TYPE,
    POSTGRES_DB_TYPE,
    MYSQL_DB_TYPE,
)
from montecarlodata.integrations.onboarding.operations.connection_ops import (
    ConnectionOperationsService,
)
from montecarlodata.integrations.onboarding.self_hosted_credentials import (
    SelfHostedCredentialOnboardingService,
)
from montecarlodata.integrations.onboarding.transactional import (
    TransactionalOnboardingService,
)
from montecarlodata.integrations.onboarding.warehouse.warehouses import (
    WarehouseOnboardingService,
)
from montecarlodata.integrations.keys import IntegrationKeyScope, IntegrationKeyService
from montecarlodata.tools import (
    add_common_options,
    validate_json_callback,
    AdvancedOptions,
    convert_uuid_callback,
    convert_empty_str_callback,
)

# Shared command verbiage
METADATA_VERBIAGE = "For metadata"
QL_VERBIAGE = "For query logs"
SQL_VERBIAGE = "For health queries"
EVENT_VERBIAGE = "For tracking data freshness and volume at scale. Requires s3 notifications to be configured first"
REGION_VERBIAGE = "If not specified the region the collector is deployed in is used"
WAREHOUSE_VERBIAGE = "For metadata, query logs and metrics"
LIGHTWEIGHT_VERBIAGE = "For metadata, and custom SQL monitors"
BI_VERBIAGE = "For reports"
PASSWORD_VERBIAGE = (
    f"If you prefer a prompt (with hidden input) enter {settings.SHOW_PROMPT_VALUE}"
)
RESOURCE_VERBIAGE = "This can be helpful if the resources are in different accounts"
CONNECTION_ID_VERBIAGE = "ID for the connection."


# Options shared across commands
def role_options(required: bool = True) -> List:
    return [
        click.option(
            "--role",
            help="Assumable role ARN to use for accessing AWS resources.",
            required=required,
        ),
        click.option(
            "--external-id",
            help="An external id, per assumable role conditions.",
            required=False,
        ),
    ]


# Name is used for the creation of a warehouse that will contain the connection.
def warehouse_create_option(required: bool = False, default: str = None) -> List:
    return _warehouse_option(
        help="Friendly name for the created integration (e.g. warehouse). Name must be unique.",
        required=required,
        default=default,
    )


# Name is used to identify an existing warehouse that will contain the connection.
def warehouse_select_option(required: bool = False, default: str = None) -> List:
    return _warehouse_option(
        help="Friendly name of the warehouse which the connection will belong to.",
        required=required,
        default=default,
    )


def _warehouse_option(help: str, required: bool = False, default: str = None) -> List:
    return [click.option("--name", help=help, required=required, default=default)]


# Name is used to give a distinctive name to the etl connection.
def etl_select_option(required: bool = False, default: str = None) -> List:
    return _etl_option(
        help="Friendly name for the etl connection.",
        required=required,
        default=default,
    )


def _etl_option(help: str, required: bool = False, default: str = None) -> List:
    return [click.option("--name", help=help, required=required, default=default)]


def port_create_option(default_port: int) -> List:
    return [
        click.option(
            "--port",
            help="HTTP port.",
            default=default_port,
            type=click.INT,
            show_default=True,
        )
    ]


WAREHOUSE_OPTIONAL_CREATE_OPTIONS = [
    click.option(
        "--create-warehouse",
        help="Create a new warehouse with this connection",
        type=click.BOOL,
        default=True,
        required=False,
    )
]
S3_OPTIONS = [
    click.option(
        "--bucket", help="S3 Bucket where query logs are contained.", required=True
    ),
    click.option("--prefix", help="Path to query logs.", required=True),
    *role_options(required=False),
]

BASIC_DATABASE_OPTIONS = [
    click.option("--host", help="Hostname.", required=True),
    click.option("--user", help="Username with access to the database.", required=True),
    click.option(
        "--password",
        help=f"User's password. {PASSWORD_VERBIAGE}.",
        required=True,
        cls=AdvancedOptions,
        prompt_if_requested=True,
    ),
]

DATABASE_OPTIONS = [
    *BASIC_DATABASE_OPTIONS,
    click.option("--database", help="Name of database.", required=True),
]

CONNECTION_OPTIONS = [
    click.option(
        "--connection-id",
        help=CONNECTION_ID_VERBIAGE,
        required=True,
        type=click.UUID,
        callback=convert_uuid_callback,
    )
]

ONBOARDING_CONFIGURATION_OPTIONS = [
    *DISAMBIGUATE_DC_OPTIONS,
    click.option(
        "--skip-validation",
        is_flag=True,
        help="Skip all connection tests.",
        cls=AdvancedOptions,
        mutually_exclusive_options=["validate_only"],
    ),
    click.option(
        "--validate-only",
        is_flag=True,
        help="Run connection tests without adding.",
        cls=AdvancedOptions,
        mutually_exclusive_options=["skip_validation"],
    ),
    click.option(
        "--auto-yes",
        is_flag=True,
        help="Skip any interactive approval.",
        default=False,
        show_default=True,
    ),
]

NETWORK_OPTIONS = [
    click.option(
        "--test-network-only",
        "skip_permission_tests",
        is_flag=True,
        default=False,
        show_default=True,
        cls=AdvancedOptions,
        mutually_exclusive_options=["skip_validation"],
        help="Skip any permission tests. Only validates network connection between the collector and resource can be established.",
    )
]

BI_OPTIONS = [
    click.option(
        "--verify-ssl/--no-verify-ssl",
        "verify_ssl",
        required=False,
        default=True,
        show_default=True,
        help="Whether to verify the SSL connection (uncheck for self-signed certs).",
    )
]

BASE_DATABRICKS_OPTIONS = [
    click.option(
        "--databricks-workspace-url", help="Databricks workspace URL.", required=True
    ),
    click.option(
        "--databricks-token",
        help=f"Databricks access token. {PASSWORD_VERBIAGE}.",
        required=True,
        cls=AdvancedOptions,
        prompt_if_requested=True,
    ),
]

DATABRICKS_OPTIONS = [
    *BASE_DATABRICKS_OPTIONS,
    click.option(
        "--databricks-workspace-id", help="Databricks workspace ID.", required=True
    ),
    click.option(
        "--databricks-cluster-id", help="Databricks cluster ID.", required=True
    ),
]

DATABRICKS_SQL_WAREHOUSE_OPTIONS = [
    *BASE_DATABRICKS_OPTIONS,
    click.option(
        "--databricks-warehouse-id", help="Databricks warehouse ID.", required=True
    ),
]

DATABRICKS_DATA_COLLECTOR_OPTIONS = [
    click.option(
        "--skip-secret-creation",
        help="Skip secret creation. Warning: Advanced use-case only.",
        default=False,
        show_default=True,
        is_flag=True,
    ),
    click.option(
        "--databricks-secret-key",
        help="Databricks secret key. This is a secret used by Databricks to communicate back to Monte Carlo.",
        default=DEFAULT_SECRET_NAME,
        show_default=True,
    ),
    click.option(
        "--databricks-secret-scope",
        help="Databricks secret scope. This is a secret used by Databricks to communicate back to Monte Carlo.",
        default=DEFAULT_SECRET_SCOPE,
        show_default=True,
    ),
    click.option(
        "--skip-notebook-creation",
        help="Skip notebook creation.",
        default=False,
        show_default=True,
        is_flag=True,
        cls=AdvancedOptions,
        required_with_options=[
            "databricks_job_id",
            "databricks_job_name",
            "databricks_notebook_path",
        ],
    ),
    click.option(
        "--databricks-job-id",
        help="Databricks job id, required if notebook creation is skipped.",
        required=False,
        cls=AdvancedOptions,
        required_with_options=["skip_notebook_creation"],
    ),
    click.option(
        "--databricks-job-name",
        help="Databricks job name, required if notebook creation is skipped.",
        required=False,
        cls=AdvancedOptions,
        required_with_options=["skip_notebook_creation"],
    ),
    click.option(
        "--databricks-notebook-path",
        help="Databricks notebook path, required if notebook creation is skipped.",
        required=False,
        cls=AdvancedOptions,
        required_with_options=["skip_notebook_creation"],
    ),
    click.option(
        "--databricks-notebook-source",
        help='Databricks notebook source, required if notebook creation is skipped. (e.g. "resources/databricks/notebook/v1/collection.py")',
        required=False,
        cls=AdvancedOptions,
        required_with_options=["skip_notebook_creation"],
    ),
]

S3_BUCKET_OPTIONS = [
    click.option(
        "--bucket",
        "bucket_name",
        required=False,
        help="Name of bucket to enable events for.",
    ),
    click.option(
        "--prefix",
        default=None,
        callback=convert_empty_str_callback,
        required=False,
        help="Limit the notifications to objects starting with a prefix (e.g. 'data/').",
    ),
    click.option(
        "--suffix",
        default=None,
        callback=convert_empty_str_callback,
        required=False,
        help="Limit notifications to objects ending with a suffix (e.g. '.csv').",
    ),
    click.option(
        "--topic-arn",
        default=None,
        callback=convert_empty_str_callback,
        required=False,
        help="Use an existing SNS topic (same region as the bucket). Creates a topic if one is not specified or "
        "if an MCD topic does not already exist in the region.",
    ),
    click.option(
        "--buckets-filename",
        required=False,
        help="Filename that contains bucket config to enable events for",
        cls=AdvancedOptions,
        mutually_exclusive_options=["bucket_name"],
    ),
]

EVENT_TYPE_OPTIONS = [
    click.option(
        "--event-type",
        required=True,
        default=CloudResourceService.MCD_EVENT_TYPE_MD,
        show_default=True,
        help="Type of event to setup.",
        type=click.Choice(
            CloudResourceService.MCD_EVENT_TYPE_FRIENDLY_TO_COLLECTOR_OUTPUT_MAP.keys(),
            case_sensitive=True,
        ),
    ),
]

RESOURCE_PROFILE_OPTIONS = [
    click.option(
        "--resource-aws-profile",
        "bucket_aws_profile",
        required=False,
        help=f"Override the AWS profile use by the CLI for operations on S3/SNS. {RESOURCE_VERBIAGE}.",
    )
]

COLLECTOR_PROFILE_OPTIONS = [
    click.option(
        "--collector-aws-profile",
        required=False,
        help=f"Override the AWS profile use by the CLI for operations on SQS/Collector. {RESOURCE_VERBIAGE}.",
    )
]

AUTO_YES_OPTIONS = [
    click.option(
        "--auto-yes",
        is_flag=True,
        help="Skip any interactive approval.",
        default=False,
        show_default=True,
    )
]


@click.group(help="Manage or integrate an asset with Monte Carlo.")
def integrations():
    """
    Group for any integration related subcommands
    """
    pass


@integrations.command(
    help=f"Setup a Hive metastore integration (MySQL). {METADATA_VERBIAGE}."
)
@click.pass_obj
@click.option(
    "--port", help="HTTP port.", default=3306, type=click.INT, show_default=True
)
@click.option(
    "--use-ssl",
    help="Use SSL to connect (using AWS RDS certificates).",
    required=False,
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--catalog",
    help="Presto catalog name. For using multiple hive clusters with Presto. "
    "Uses 'hive' if not specified.",
    required=False,
)
@add_common_options(warehouse_create_option())
@add_common_options(WAREHOUSE_OPTIONAL_CREATE_OPTIONS)
@add_common_options(DATABASE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_hive_metastore(ctx, database, name, **kwargs):
    """
    Onboard a hive metastore connection (MySQL)
    """
    HiveOnboardingService(config=ctx["config"]).onboard_hive_mysql(
        dbName=database, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a Presto SQL integration. {SQL_VERBIAGE}.")
@click.pass_obj
@click.option("--host", help="Hostname.", required=True)
@click.option(
    "--port", help="HTTP port.", default=8889, type=click.INT, show_default=True
)
@click.option("--user", help="Username with access to catalog/schema.", required=False)
@click.option(
    "--password",
    help=f"User's password. {PASSWORD_VERBIAGE}",
    required=False,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option("--catalog", help="Mount point to access data source.", required=False)
@click.option("--schema", help="Schema to access.", required=False)
@click.option(
    "--http-scheme",
    help="Scheme for authentication.",
    type=click.Choice(["http", "https"], case_sensitive=True),
    required=True,
)
@click.option(
    "--cert-file",
    help="Local SSL certificate file to upload to collector.",
    required=False,
    type=click.Path(dir_okay=False, exists=True),
    cls=AdvancedOptions,
    mutually_exclusive_options=["cert_s3"],
)
@click.option(
    "--cert-s3",
    help="Object path (key) to a certificate already uploaded to the collector.",
    required=False,
    cls=AdvancedOptions,
    mutually_exclusive_options=["cert_file"],
)
@click.option(
    "--skip-cert-verification",
    help="Skip SSL certificate verification.",
    required=False,
    is_flag=True,
)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_presto(ctx, password, name, **kwargs):
    """
    Onboard a presto sql connection
    """
    if not password:
        password = None  # make explicitly null if not set. Prompts can't be None
    PrestoOnboardingService(config=ctx["config"]).onboard_presto_sql(
        password=password, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a Hive SQL integration. {SQL_VERBIAGE}.")
@click.pass_obj
@click.option("--host", help="Hostname.", required=True)
@click.option("--database", help="Name of database.", required=False)
@click.option(
    "--port", help="HTTP port.", default=10000, type=click.INT, show_default=True
)
@click.option("--user", help="Username with access to hive.", required=True)
@click.option(
    "--auth-mode",
    help="Hive authentication mode.",
    required=False,
    default="SASL",
    type=click.Choice(["SASL", "NOSASL"]),
    show_default=True,
)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_hive(ctx, user, name, **kwargs):
    HiveOnboardingService(config=ctx["config"]).onboard_hive_sql(
        username=user, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a Presto logs integration (S3). {QL_VERBIAGE}.")
@click.pass_obj
@add_common_options(S3_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_presto_logs(ctx, role, **kwargs):  # DEPRECATED
    """
    Onboard a presto logs (s3) connection
    """
    PrestoOnboardingService(config=ctx["config"]).onboard_presto_s3(
        assumable_role=role, **kwargs
    )


@integrations.command(help=f"Setup a Hive EMR logs integration (S3). {QL_VERBIAGE}.")
@click.pass_obj
@add_common_options(S3_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_hive_logs(ctx, role, **kwargs):  # DEPRECATED
    """
    Onboard a hive emr (s3) connection
    """
    HiveOnboardingService(config=ctx["config"]).onboard_hive_s3(
        assumable_role=role, **kwargs
    )


@integrations.command(help=f"Setup a Glue integration. {METADATA_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--region", help=f"Glue catalog region. {REGION_VERBIAGE}.", required=False
)
@add_common_options(role_options())
@add_common_options(warehouse_create_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_glue(ctx, role, region, name, **kwargs):
    """
    Onboard a glue connection
    """
    GlueAthenaOnboardingService(config=ctx["config"]).onboard_glue(
        assumable_role=role, aws_region=region, warehouseName=name, **kwargs
    )


@integrations.command(
    help="Setup an Athena integration. For query logs and health queries."
)
@click.pass_obj
@click.option(
    "--catalog",
    help="Glue data catalog. If not specified the AwsDataCatalog is used.",
    required=False,
)
@click.option(
    "--workgroup",
    help="Workbook for running queries and retrieving logs. If not specified the primary is used.",
    required=False,
)
@click.option(
    "--region", help=f"Athena cluster region. {REGION_VERBIAGE}.", required=False
)
@add_common_options(warehouse_select_option())
@add_common_options(role_options())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_athena(ctx, role, region, name, **kwargs):
    """
    Onboard an athena connection
    """
    GlueAthenaOnboardingService(config=ctx["config"]).onboard_athena(
        assumable_role=role, aws_region=region, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a thrift binary Spark integration. {SQL_VERBIAGE}.")
@click.pass_obj
@click.option("--host", help="Hostname.", required=True)
@click.option("--database", help="Name of database.", required=True)
@click.option("--port", help="Port.", default=10000, type=click.INT, show_default=True)
@click.option("--user", help="Username with access to spark.", required=True)
@click.option(
    "--password",
    help=f"User's password. {PASSWORD_VERBIAGE}.",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_spark_binary_mode(ctx, user, name, **kwargs):
    """
    Onboard a spark connection, thrift binary mode
    """
    SparkOnboardingService(config=ctx["config"]).onboard_spark(
        SPARK_BINARY_MODE_CONFIG_TYPE, username=user, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a thrift HTTP Spark integration. {SQL_VERBIAGE}.")
@click.pass_obj
@click.option("--url", help="HTTP URL.", required=True)
@click.option("--user", help="Username with access to spark.", required=True)
@click.option(
    "--password",
    help=f"User's password. {PASSWORD_VERBIAGE}.",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_spark_http_mode(ctx, user, name, **kwargs):
    """
    Onboard a spark connection, thrift http mode
    """
    SparkOnboardingService(config=ctx["config"]).onboard_spark(
        SPARK_HTTP_MODE_CONFIG_TYPE, username=user, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a Spark integration for Databricks. {SQL_VERBIAGE}.")
@click.pass_obj
@add_common_options(DATABRICKS_OPTIONS)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_spark_databricks(ctx, name, **kwargs):
    """
    Onboard a spark connection, databricks
    """
    SparkOnboardingService(config=ctx["config"]).onboard_spark(
        SPARK_DATABRICKS_CONFIG_TYPE, warehouseName=name, **kwargs
    )


@integrations.command(
    help="Setup a Databricks metastore integration. For metadata and health queries."
)
@click.pass_obj
@add_common_options(warehouse_create_option())
@add_common_options(DATABRICKS_OPTIONS)
@add_common_options(DATABRICKS_DATA_COLLECTOR_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_databricks_metastore(ctx, name, **kwargs):
    """
    Onboard a databricks metastore
    """
    DatabricksOnboardingService(
        config=ctx["config"], mc_client=create_mc_client(ctx)
    ).onboard_databricks_metastore(
        connection_type=DATABRICKS_METASTORE_CONNECTION_TYPE,
        warehouseName=name,
        **kwargs,
    )


@integrations.command(
    help="Setup a Databricks Delta integration. For metadata queries on delta tables when using an external metastore in databricks."
)
@click.pass_obj
@add_common_options(DATABRICKS_OPTIONS)
@add_common_options(warehouse_select_option())
@add_common_options(DATABRICKS_DATA_COLLECTOR_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_databricks_delta(ctx, name, **kwargs):
    """
    Onboard a databricks delta service
    """
    DatabricksOnboardingService(
        config=ctx["config"], mc_client=create_mc_client(ctx)
    ).onboard_databricks_metastore(
        connection_type=DATABRICKS_DELTA_CONNECTION_TYPE, warehouseName=name, **kwargs
    )


@integrations.command(
    help=f"Setup a Databricks SQL Warehouse integration. {SQL_VERBIAGE}"
)
@click.pass_obj
@add_common_options(DATABRICKS_SQL_WAREHOUSE_OPTIONS)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_databricks_sql_warehouse(ctx, name, **kwargs):
    """
    Onboard a databricks sql warehouse connection
    """
    DatabricksOnboardingService(
        config=ctx["config"], mc_client=create_mc_client(ctx)
    ).onboard_databricks_sql_warehouse(warehouseName=name, **kwargs)


@integrations.command(help=f"Setup a Redshift integration. {WAREHOUSE_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--port", help="HTTP port.", default=5439, type=click.INT, show_default=True
)
@add_common_options(warehouse_create_option())
@add_common_options(DATABASE_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@add_common_options(NETWORK_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_redshift(ctx, database, name, **kwargs):
    """
    Onboard a redshift connection
    """
    WarehouseOnboardingService(config=ctx["config"]).onboard_redshift(
        dbName=database, warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a Snowflake integration. {WAREHOUSE_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--user",
    help="User with access to snowflake.",
    required=True,
    cls=AdvancedOptions,
    at_least_one_set=["password", "private_key"],
)
@click.option("--account", help="Snowflake account name.", required=True)
@click.option("--warehouse", help="Name of the warehouse for the user.", required=True)
@click.option(
    "--password",
    help=f"User's password if using user/password basic auth. {PASSWORD_VERBIAGE}.",
    required=False,
    cls=AdvancedOptions,
    prompt_if_requested=True,
    mutually_exclusive_options=["private_key"],
)
@click.option(
    "--private-key",
    help="User's private key file path if using key pair auth.",
    type=click.Path(exists=True),
    required=False,
    cls=AdvancedOptions,
    mutually_exclusive_options=["password"],
)
@click.option(
    "--private-key-passphrase",
    help=f"User's private key passphrase if using key pair auth. This argument is only needed when the private key is encrypted. {PASSWORD_VERBIAGE}.",
    required=False,
    cls=AdvancedOptions,
    prompt_if_requested=True,
    mutually_exclusive_options=["password"],
    required_with_options=["private_key"],
)
@add_common_options(warehouse_create_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_snowflake(ctx, name, **kwargs):
    """
    Onboard a snowflake connection
    """
    WarehouseOnboardingService(config=ctx["config"]).onboard_snowflake(
        warehouseName=name, **kwargs
    )


@integrations.command(help=f"Setup a BigQuery integration. {WAREHOUSE_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--key-file", help="JSON Key file.", type=click.Path(exists=True), required=True
)
@add_common_options(warehouse_create_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_bigquery(ctx, key_file, name, **kwargs):
    """
    Onboard a BigQuery connection
    """
    WarehouseOnboardingService(config=ctx["config"]).onboard_bq(
        ServiceFile=key_file, warehouseName=name, **kwargs
    )


@integrations.command(help="Setup an integration that uses self-hosted credentials.")
@click.pass_obj
@click.option(
    "--connection-type",
    help="Type of connection.",
    required=True,
    type=click.Choice(GQL_TO_FRIENDLY_CONNECTION_MAP.keys(), case_sensitive=False),
    cls=AdvancedOptions,
    values_with_required_options=CONNECTION_TO_WAREHOUSE_TYPE_MAP.keys(),
    required_options_for_values=["name"],
)
@click.option(
    "--mechanism",
    help="Credential self-hosting mechanism.",
    required=True,
    type=click.Choice(SELF_HOSTING_MECHANISMS, case_sensitive=False),
    default=SECRETS_MANAGER_CREDENTIAL_MECHANISM,
)
@click.option(
    "--key",
    help="Identifier for credentials within self-hosting mechanism.",
    required=True,
)
@click.option("--name", help="Friendly name for the warehouse.", required=False)
@add_common_options(role_options(required=False))
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_self_hosted_credentials(ctx, mechanism, key, role, name, **kwargs):
    """
    Onboard a connection with self-hosted credentials
    """
    SelfHostedCredentialOnboardingService(config=ctx["config"]).onboard_connection(
        self_hosting_mechanism=mechanism,
        self_hosting_key=key,
        assumable_role=role,
        warehouseName=name,
        **kwargs,
    )


@integrations.command(
    help=f"Setup an Airflow integration to receive events from Airflow callbacks."
)
@click.pass_obj
@click.option("--host", help="Hostname.", required=True)
@add_common_options(etl_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_airflow(ctx, host, name, **kwargs):
    """
    Onboard an airflow connection
    """
    etl_name = name or "airflow"
    AirflowOnboardingService(config=ctx["config"]).onboard_airflow(
        etl_name=etl_name, hostName=host, **kwargs
    )


@integrations.command(help="Configure S3 events for Airflow task logs.")
@click.pass_obj
@add_common_options(warehouse_create_option(default="airflow-logs-s3"))
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@add_common_options(role_options())
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def configure_airflow_log_events(ctx, **kwargs):
    """
    Configure s3 events for Airflow task logs
    """
    EventsOnboardingService(config=ctx["config"]).configure_airflow_events(**kwargs)


@integrations.command(help=f"Configure S3 metadata events. {EVENT_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--connection-type",
    help="Type of the integration.",
    cls=AdvancedOptions,
    mutually_exclusive_options=["connection_id"],
    type=click.Choice(
        [
            DATABRICKS_METASTORE_CONNECTION_TYPE,
            GLUE_CONNECTION_TYPE,
            HIVE_MYSQL_CONNECTION_TYPE,
        ],
        case_sensitive=False,
    ),
    required=True,
)
@add_common_options(warehouse_create_option(default="s3-metadata-events"))
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def configure_metadata_events(ctx, **kwargs):
    """
    Configure S3 metadata events for a lake
    """
    EventsOnboardingService(config=ctx["config"]).configure_metadata_events(**kwargs)


@integrations.command(help=f"Configure S3 query log events. {EVENT_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--connection-type",
    help="Type of the integration.",
    type=click.Choice(
        [HIVE_S3_CONNECTION_TYPE, PRESTO_S3_CONNECTION_TYPE], case_sensitive=True
    ),
    required=True,
)
@add_common_options(warehouse_select_option())
@add_common_options(role_options())
@click.option(
    "--format-type",
    help="Query log format.",
    type=click.Choice(["hive-emr", "hive-native", "custom"], case_sensitive=True),
    required=True,
)
@click.option(
    "--source-format",
    help='Query log file format. Only required when "custom" is used.',
    type=click.Choice(["json", "jsonl"], case_sensitive=True),
    required=False,
)
@click.option(
    "--mapping-file",
    help='Mapping of expected to existing query log fields. Only required if "custom" is used.',
    type=click.Path(exists=True),
    required=False,
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def configure_query_log_events(ctx, **kwargs):
    """
    Configure S3 query log events for a lake
    """
    EventsOnboardingService(config=ctx["config"]).configure_query_log_events(**kwargs)


@integrations.command(help=f"Disable S3 events for Airflow task logs.")
@click.pass_obj
@click.option("--name", help="Resource name (only required if more than one exists)")
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def disable_airflow_log_events(ctx, **kwargs):
    """
    Disable S3 events for Airflow task logs
    """
    EventsOnboardingService(config=ctx["config"]).disable_airflow_log_events(**kwargs)


@integrations.command(help=f"Disable S3 metadata events.")
@click.pass_obj
@click.option("--name", help="Resource name (only required if more than one exists)")
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def disable_metadata_events(ctx, **kwargs):
    """
    Configure S3 metadata events
    """
    EventsOnboardingService(config=ctx["config"]).disable_metadata_events(**kwargs)


@integrations.command(help=f"Disable S3 query log events.")
@click.pass_obj
@click.option("--name", help="Resource name (only required if more than one exists)")
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def disable_query_log_events(ctx, **kwargs):
    """
    Configure S3 query log events
    """
    EventsOnboardingService(config=ctx["config"]).disable_query_log_events(**kwargs)


@integrations.command(help=f"Setup a Tableau integration. {BI_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--token-name",
    help="Name for the personal access token.",
    cls=AdvancedOptions,
    mutually_exclusive_options=["password", "username"],
    required_with_options=["token_value"],
    at_least_one_set=["password", "username", "token_name", "token_value"],
)
@click.option(
    "--token-value",
    help=f"Value for the personal access token. {PASSWORD_VERBIAGE}.",
    cls=AdvancedOptions,
    mutually_exclusive_options=["password", "username"],
    required_with_options=["token_name"],
    prompt_if_requested=True,
)
@click.option(
    "--password",
    help=f"Password for the service account. {PASSWORD_VERBIAGE}.",
    cls=AdvancedOptions,
    mutually_exclusive_options=["token_name", "token_value"],
    required_with_options=["user"],
    prompt_if_requested=True,
)
@click.option(
    "--user",
    help="Username for the service account.",
    cls=AdvancedOptions,
    mutually_exclusive_options=["token_name", "token_value"],
    required_with_options=["password"],
)
@click.option("--site-name", help="The Tableau site name.", required=True)
@click.option("--server-name", help="The Tableau server name.", required=True)
@click.option(
    "--name", help="Friendly name for the Tableau integration.", required=False
)
@add_common_options(BI_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_tableau(ctx, user, **kwargs):
    """
    Onboard a tableau connection
    """
    ReportsOnboardingService(config=ctx["config"]).onboard_tableau(
        username=user, **kwargs
    )


@integrations.command(help=f"Setup a Looker metadata integration. {BI_VERBIAGE}.")
@click.pass_obj
@click.option("--host-url", "base_url", help="Looker host url.", required=True)
@click.option("--client-id", help="Looker client id.", required=True)
@click.option(
    "--client-secret",
    help=f"Looker client secret (API key). {PASSWORD_VERBIAGE}.",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option(
    "--name", help="Friendly name for the Looker integration.", required=False
)
@add_common_options(BI_OPTIONS)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_looker(ctx, **kwargs):
    """
    Onboard a looker metadata connection
    """
    ReportsOnboardingService(config=ctx["config"]).onboard_looker_metadata(**kwargs)


@integrations.command(
    help="""
Create or refresh details of BI warehouse sources. Warehouse sources are warehouses connected to a BI container. 

If only the bi-container-id parameter is supplied, then the behavior will be the following: 
\b
\n
(1) All warehouse sources in the customer's BI system will be matched against warehouses in Monte Carlo. 
\b
\n
(2) If all of the customer's BI warehouse sources have a match in Monte Carlo, then the warehouse source details will be
saved.
\b
\n
(3) Otherwise, the response will contain the details of customer's BI warehouse sources and Monte Carlo warehouses. The 
caller will then need to manually save the warehouse source details by specifying the warehouse-source-details
parameter.
\b
\n
If the warehouse-source-details parameter is also supplied, then the warehouses will be validated and saved.
"""
)
@click.pass_obj
@click.option(
    "--bi-container-id",
    help="ID of the BI container for which to refresh warehouse sources.",
    required=True,
)
@click.option(
    "--warehouse-source-details",
    help="""
              An optional JSON array of warehouse sources. If supplied, these warehouse sources will be created. If an
              entry with the same bi_container_id, warehouse_resource_id and warehouse_resource_type is found, its
              bi_warehouse_id will be updated.
              
              Each JSON object in this array has the following fields:
              \b
              \n
              bi_warehouse_id: The ID of the warehouse in the customer's ID space.
              \b
              \n
              warehouse_resource_id: The ID of the warehouse Monte Carlo's ID space.
              \b
              \n
              warehouse_resource_type: The type of the warehouse.
              \b
              \n
              E.g. --warehouse-source-details '[{"bi_warehouse_id":"0dd5d40e-8749-71cb-9fa5-2e33b570ff43", 
              "warehouse_resource_id": "11b1ad6f-e35c-4532-84d8-2fc88bd53660", "warehouse_resource_type": "snowflake"}]'
              """,
    required=False,
    callback=validate_json_callback,
)
def refresh_bi_to_warehouse_connections(
    ctx, bi_container_id, warehouse_source_details, **kwargs
):
    BiService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
    ).refresh_bi_to_warehouse_connections(bi_container_id, warehouse_source_details)


@integrations.command(help=f"Setup a Looker ML (git) integration. {BI_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--ssh-key",
    help="The ssh key for git ssh integrations.",
    required=False,
    type=click.Path(dir_okay=False, exists=True),
    cls=AdvancedOptions,
    mutually_exclusive_options=["token", "username"],
)
@click.option(
    "--repo-url",
    help="Repository URL as ssh://[user@]server/project.git or the shorter "
    "form [user@]server:project.git for ssh. For https, use https://server/project.git.",
    required=True,
)
@click.option(
    "--token",
    help="Git Access Token to be used for Https instead of ssh key.",
    required=False,
    cls=AdvancedOptions,
    mutually_exclusive_options=["ssh-key"],
)
@click.option(
    "--username",
    help="Git username to be used in conjunction with the access token. This is only "
    "required for BitBucket integrations.",
    cls=AdvancedOptions,
    required=False,
    mutually_exclusive_options=["ssh-key"],
)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_looker_git(ctx, **kwargs):
    """
    Onboard a looker metadata connection
    """
    ReportsOnboardingService(config=ctx["config"]).onboard_looker_git(**kwargs)


@integrations.command(help=f"Setup a Power BI integration. {BI_VERBIAGE}.")
@click.pass_obj
@click.option(
    "--tenant-id", help="The tenant ID from the Azure Active Directory.", required=True
)
@click.option(
    "--auth-mode",
    help="Authentication Mode. We support service principal and primary user two auth types",
    required=True,
    type=click.Choice(["SERVICE_PRINCIPAL", "PRIMARY_USER"]),
)
@click.option(
    "--client-id",
    help="App registration application ID for accessing Power BI.",
    required=True,
    cls=AdvancedOptions,
)
@click.option(
    "--client-secret",
    help="Secret for the application to access the Power BI. Set only when auth-mode is "
    "SERVICE_PRINCIPAL",
    required=False,
    cls=AdvancedOptions,
    mutually_exclusive_options=["username", "password"],
    prompt_if_requested=True,
)
@click.option(
    "--username",
    help="Username for accessing the Power BI. Set only when auth-mode is PRIMARY_USER.",
    cls=AdvancedOptions,
    required=False,
    mutually_exclusive_options=["client-secret"],
)
@click.option(
    "--password",
    help="Password for accessing the Power BI. Set only when auth-mode is PRIMARY_USER.",
    cls=AdvancedOptions,
    required=False,
    mutually_exclusive_options=["client-secret"],
    prompt_if_requested=True,
)
@click.option(
    "--name", help="Friendly name for the Power BI integration.", required=False
)
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_power_bi(ctx, **kwargs):
    """
    Onboard a Power BI connection
    """
    ReportsOnboardingService(config=ctx["config"]).onboard_power_bi(**kwargs)


@integrations.command(help="Setup a dbt Cloud integration.")
@click.pass_obj
@click.option(
    "--dbt-cloud-api-token",
    "dbt_cloud_api_token",
    help=f"dbt Cloud API token. {PASSWORD_VERBIAGE}.",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option(
    "--dbt-cloud-account-id",
    "dbt_cloud_account_id",
    help="dbt Cloud Account ID.",
    required=True,
)
@click.option(
    "--dbt-cloud-base-url",
    "dbt_cloud_base_url",
    help="dbt Cloud Base URL.",
    required=False,
)
@click.option(
    "--webhook-hmac-secret",
    "webhook_hmac_secret",
    help="The secret token `hmac_secret` provided by dbt after a webhook is successfully created in dbt.",
    required=False,
)
@click.option(
    "--webhook-id",
    "webhook_id",
    help="The webhook id provided by dbt after a webhook is successfully created in dbt.",
    required=False,
)
@add_common_options(warehouse_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_dbt_cloud(ctx, name, **kwargs):
    """
    Onboard a dbt cloud connection
    """
    DbtCloudOnboardingService(config=ctx["config"]).onboard_dbt_cloud(
        warehouseName=name, **kwargs
    )


@integrations.command(help="Setup a Fivetran integration.")
@click.pass_obj
@click.option(
    "--fivetran-api-key",
    "fivetran_api_key",
    help=f"Fivetran API Key. {PASSWORD_VERBIAGE}.",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option(
    "--fivetran-api-password",
    "fivetran_api_password",
    help=f"Fivetran API Key. {PASSWORD_VERBIAGE}.",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option(
    "--fivetran-base-url",
    "fivetran_base_url",
    help="Fivetran Base URL.",
    required=False,
)
@add_common_options(etl_select_option())
@add_common_options(ONBOARDING_CONFIGURATION_OPTIONS)
def add_fivetran(ctx, name, **kwargs):
    """
    Onboard a Fivetran connection
    """
    FivetranOnboardingService(config=ctx["config"]).onboard_fivetran(
        etl_name=name, **kwargs
    )


@integrations.command(help=f"Setup a SQL Server integration. {LIGHTWEIGHT_VERBIAGE}.")
@click.pass_obj
@add_common_options(
    warehouse_create_option(required=True)
    + port_create_option(default_port=1433)
    + BASIC_DATABASE_OPTIONS
    + ONBOARDING_CONFIGURATION_OPTIONS
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_sql_server(ctx, name, **kwargs):
    """
    Onboard a SQL Server connection
    """
    TransactionalOnboardingService(config=ctx["config"]).onboard_transactional_db(
        warehouseName=name, dbType=SQL_SERVER_DB_TYPE, **kwargs
    )


@integrations.command(help=f"Setup a Postgres integration. {LIGHTWEIGHT_VERBIAGE}.")
@click.pass_obj
@add_common_options(
    warehouse_create_option(required=True)
    + port_create_option(default_port=5432)
    + DATABASE_OPTIONS
    + ONBOARDING_CONFIGURATION_OPTIONS
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_postgres(ctx, database, name, **kwargs):
    """
    Onboard a Postgres connection
    """
    TransactionalOnboardingService(config=ctx["config"]).onboard_transactional_db(
        warehouseName=name, dbName=database, dbType=POSTGRES_DB_TYPE, **kwargs
    )


@integrations.command(help=f"Setup a MySQL integration. {LIGHTWEIGHT_VERBIAGE}.")
@click.pass_obj
@add_common_options(
    warehouse_create_option(required=True)
    + port_create_option(default_port=3306)
    + BASIC_DATABASE_OPTIONS
    + ONBOARDING_CONFIGURATION_OPTIONS
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_mysql(ctx, name, **kwargs):
    """
    Onboard a MySQL connection
    """
    TransactionalOnboardingService(config=ctx["config"]).onboard_transactional_db(
        warehouseName=name, dbType=MYSQL_DB_TYPE, **kwargs
    )


@integrations.command(help="List all active connections.", name="list")
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def display_integrations(ctx):
    """
    Display active integrations
    """
    OnboardingStatusService(config=ctx["config"]).display_integrations()


@integrations.command(
    help="Create an IAM role from a policy FILE. "
    "The returned role ARN and external id should be used for adding lake assets."
)
@click.pass_obj
@click.argument("file", type=click.Path(dir_okay=False, exists=True))
@click.option(
    "--aws-profile",
    required=False,
    help="Override the AWS profile used by the CLI, which determines where the role is created. "
    "This can be helpful when the account that manages the asset is not the same as the collector.",
)
def create_role(ctx, file, aws_profile):
    """
    Create a collector compatible role from the provided policy
    """
    CloudResourceService(
        config=ctx["config"], aws_profile_override=aws_profile
    ).create_role(path_to_policy_doc=file)


@integrations.command(
    help="Update credentials for a connection. Only replaces/inserts the keys in changes by default.",
    name="update",
)
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
@click.option(
    "--changes",
    help="""
              Credential key,value pairs as JSON.
              \b
              \n
              E.g. --changes '{"user":"Apollo"}'
              """,
    required=True,
    callback=validate_json_callback,
)
@click.option(
    "--skip-validation",
    help="Skip validating credentials.",
    default=False,
    show_default=True,
    is_flag=True,
)
@click.option(
    "--replace-all",
    help="Replace all credentials rather than just inserting/updating the keys in changes.",
    default=False,
    show_default=True,
    is_flag=True,
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def update_credentials(ctx, connection_id, changes, skip_validation, replace_all):
    """
    Update credentials for a connection
    """
    ConnectionOperationsService(config=ctx["config"]).update_credentials(
        connection_id=connection_id,
        changes=changes,
        should_validate=not skip_validation,
        should_replace=replace_all,
    )


@integrations.command(
    help="Remove an existing connection. Deletes any associated jobs, monitors, etc.",
    name="remove",
)
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
@click.option(
    "--no-prompt",
    help="Don't ask for confirmation.",
    default=False,
    show_default=True,
    is_flag=True,
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def remove_connection(ctx, **kwargs):
    """
    Remove connection by ID
    """
    ConnectionOperationsService(config=ctx["config"]).remove_connection(**kwargs)


@integrations.command(help="Retest an existing connection.", name="test")
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def echo_test_existing(ctx, **kwargs):
    """
    Tests an existing connection and echos results in pretty JSON.
    """
    ConnectionOperationsService(config=ctx["config"]).echo_test_existing(**kwargs)


@integrations.command(help="Setup complete S3 event notifications for a lake.")
@click.pass_obj
@add_common_options(S3_BUCKET_OPTIONS)
@add_common_options(EVENT_TYPE_OPTIONS)
@add_common_options(COLLECTOR_PROFILE_OPTIONS)
@add_common_options(RESOURCE_PROFILE_OPTIONS)
@add_common_options(AUTO_YES_OPTIONS)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_events(ctx, **kwargs):
    CloudResourceService(config=ctx["config"], aws_wrapper=None).add_events(**kwargs)


@integrations.command(help="Setup Event Topic for S3 event notifications in a lake.")
@click.pass_obj
@add_common_options(S3_BUCKET_OPTIONS)
@add_common_options(EVENT_TYPE_OPTIONS)
@add_common_options(RESOURCE_PROFILE_OPTIONS)
@add_common_options(AUTO_YES_OPTIONS)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def create_event_topic(ctx, **kwargs):
    CloudResourceService(config=ctx["config"], aws_wrapper=None).create_event_topic(
        **kwargs
    )


@integrations.command(help="Setup Bucket Side S3 event infrastructure for a lake.")
@click.pass_obj
@add_common_options(S3_BUCKET_OPTIONS)
@add_common_options(EVENT_TYPE_OPTIONS)
@add_common_options(RESOURCE_PROFILE_OPTIONS)
@add_common_options(AUTO_YES_OPTIONS)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def create_bucket_side_event_infrastructure(ctx, **kwargs):
    CloudResourceService(
        config=ctx["config"], aws_wrapper=None
    ).create_bucket_side_event_infrastructure(**kwargs)


@integrations.command(
    help="Create an integration key. The resulting key id and secret will be printed to the console."
)
@click.pass_obj
@click.option("--description", required=True, help="Key description.")
@click.option(
    "--scope",
    required=True,
    type=click.Choice(IntegrationKeyScope.values(), case_sensitive=False),
    help="Key scope (integration the key can be used for).",
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def create_key(ctx, **kwargs):
    IntegrationKeyService(config=ctx["config"]).create(**kwargs)


@integrations.command(help="Delete an integration key.")
@click.pass_obj
@click.option("--key-id", required=True, help="Integration key id.")
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def delete_key(ctx, **kwargs):
    IntegrationKeyService(config=ctx["config"]).delete(**kwargs)


@integrations.command(help="List all integration keys.")
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def list_keys(ctx):
    IntegrationKeyService(config=ctx["config"]).get_all()


@integrations.command(help="Set the name of a Warehouse.")
@click.pass_obj
@click.option("--current-name", required=True, help="Current Name of the Warehouse")
@click.option("--new-name", required=True, help="Name to give the Warehouse")
def set_warehouse_name(ctx, **kwargs):
    ConnectionOperationsService(config=ctx["config"]).set_warehouse_name(**kwargs)


@integrations.command(help="Set the name of a BI Connection.")
@click.pass_obj
@click.option(
    "--bi-connection-id", required=True, help="UUID of the existing BI Connection"
)
@click.option("--new-name", required=True, help="Name to give the BI Connection")
def set_bi_connection_name(ctx, **kwargs):
    ConnectionOperationsService(config=ctx["config"]).set_bi_connection_name(**kwargs)


@integrations.command(help="Update Databricks Notebook to the latest version.")
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
def update_databricks_notebook(ctx, **kwargs):
    DatabricksOnboardingService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
    ).update_databricks_notebook(**kwargs)


@integrations.command(help="Get the Databricks job info for your connection.")
@click.pass_obj
@add_common_options(CONNECTION_OPTIONS)
def show_databricks_metadata_job_info(ctx, **kwargs):
    DatabricksOnboardingService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
    ).get_databricks_job_info(**kwargs)


@integrations.command(help="Get the most up to date databricks notebook version.")
@click.pass_obj
def show_current_notebook_version(ctx):
    DatabricksOnboardingService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
    ).get_current_databricks_notebook_version()



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
    EXPECTED_TEST_AIRFLOW_RESPONSE_FIELD,
    AIRFLOW_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    ADD_ETL_CONNECTION_MUTATION,
    TEST_AIRFLOW_CRED_MUTATION,
)


class AirflowOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_airflow(self, **kwargs) -> None:
        self.onboard(
            validation_query=TEST_AIRFLOW_CRED_MUTATION,
            validation_response=EXPECTED_TEST_AIRFLOW_RESPONSE_FIELD,
            connection_query=ADD_ETL_CONNECTION_MUTATION,
            connection_response=EXPECTED_ADD_ETL_CONNECTION_RESPONSE_FIELD,
            connection_type=AIRFLOW_CONNECTION_TYPE,
            **kwargs
        )



import contextlib
import fnmatch
import glob
import json
import os
import re
from datetime import datetime, timezone
from functools import reduce
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import click
import yaml
from box import Box
from marshmallow import ValidationError
from pycarlo.core import Client, Query
from tabulate import tabulate

from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.fs_utils import mkdirs
from montecarlodata.iac.client import MonteCarloConfigTemplateClient
from montecarlodata.iac.schemas import (
    ProjectConfig,
    ResourceModification,
    ConfigTemplateUpdateState,
)
from montecarlodata.iac.utils import (
    has_montecarlo_property,
    is_dbt_schema,
    has_namespace_property,
)
from montecarlodata.settings import PROJECT_CONFIG_FILENAME, TARGET_DIRECTORY_NAME
from montecarlodata.utils import GqlWrapper


class MonteCarloConfigService:
    NS_HEADERS = ("Namespace", "Last Update Time")
    MORE_NS_MESSAGE = (
        "There are more namespaces available. Increase the limit to view them."
    )
    DBT_REF_PATTERN = re.compile(r"\[\[\s*ref\(['\"](\w+)['\"]\)\s*\]\]")
    SQL_KEYS = {"sql", "sampling_sql", "variables"}

    def __init__(
        self,
        config: Config,
        pycarlo_client: Client,
        request_wrapper: Optional[GqlWrapper] = None,
        project_dir: Optional[str] = None,
        dbt_manifest_path: Optional[str] = None,
        print_func: Optional[Callable] = None,
    ):
        request_wrapper = request_wrapper or GqlWrapper(config)
        self.client = MonteCarloConfigTemplateClient(request_wrapper)
        self._pycarlo_client = pycarlo_client
        self._print_func = print_func or click.echo
        self.project_dir = os.path.abspath(project_dir) if project_dir else os.getcwd()
        self.dbt_manifest_path = (
            os.path.abspath(dbt_manifest_path) if dbt_manifest_path else None
        )
        self._abort_on_error = True

        # Nicer formatting for multiline strings
        yaml.add_representer(str, self._str_presenter)

    def _init_local_config(self):
        self.target_dir = self._create_target_directory()
        self.project_config = self._load_project_config()

    def _init_dbt_config(self):
        self.dbt_project = self._get_dbt_project_name()
        self.dbt_nodes = self._load_dbt_manifest_nodes()

    @manage_errors
    def apply(
        self,
        apply_namespace: str,
        dry_run: bool = False,
        skip_confirmation: bool = False,
        abort_on_error=True,
    ):
        """
        Compile configuration.
        Submit configuration via API
        """
        _, ns_mc_config_dict, _ = self.compile(apply_namespace, abort_on_error)
        assert self.project_config is not None
        responses = []
        for namespace, mc_config_dict in ns_mc_config_dict.items():
            self._print_func(
                "----------------------------------------------------------------"
            )
            self._print_func(f"Computing changes for namespace: {namespace}")
            self._print_func(
                "----------------------------------------------------------------"
            )
            response = self.client.apply_config_template(
                namespace,
                mc_config_dict,
                resource=self.project_config.default_resource,
                dry_run=True,
                misconfigured_as_warning=True,
            )

            self._check_apply_errors(response, abort_on_error)

            if response.warnings:
                click.echo()
                misconfigured_warnings = response.warnings.get(
                    "misconfigured_warnings", []
                )
                if misconfigured_warnings:
                    click.echo(
                        "Warnings encountered, please review to avoid "
                        "creating misconfigured monitors:"
                    )
                    self._echo_misconfigured_warnings(misconfigured_warnings)

            if response.resource_modifications:
                self._print_func()
                self._print_func("Modifications:")
                for resource_modification in response.resource_modifications:
                    self._print_func(
                        f' - {"[DRY RUN] " if dry_run else ""}{resource_modification.type} '
                        f"- {resource_modification.description}"
                        f"{self._generate_significance_text(resource_modification)}"
                    )
                self._print_func()

                if dry_run:
                    self._print_func("Dry run, none of the changes have been applied.")
                    self._print_func()
                elif skip_confirmation or click.confirm(
                    "Do you want to apply this change?", abort=False
                ):
                    response = self.client.apply_config_template(
                        namespace,
                        mc_config_dict,
                        resource=self.project_config.default_resource,
                        dry_run=False,
                        misconfigured_as_warning=True,
                    )
                    success = self._check_apply_errors(response, abort_on_error)

                    if success:
                        self._print_func("Changes successfully applied.")
                    else:
                        self._print_func("Failed to apply changes.")
                    self._print_func()
                else:
                    self._print_func("None of the changes have been applied.")
            else:
                self._print_func()
                self._print_func("No changes to configuration found, doing nothing.")
                self._print_func()
            responses.append(response)
        return responses

    def _check_apply_errors(
        self,
        response: ConfigTemplateUpdateState,
        abort_on_error: bool,
    ) -> bool:
        if response.state == "SKIPPED":
            self._print_func()
            self._print_func("Update skipped in favor of more recent update.")

            if abort_on_error:
                complain_and_abort("Update skipped, exiting.")

            return False

        if not response.errors:
            return True

        self._print_func()
        self._print_func("Errors encountered when attempting to apply configuration:")

        system_error = response.errors.get("system_error", {})
        if system_error:
            self._print_func(f" - {system_error}")
        else:
            validation_errors = response.errors.get("validation_errors", {})
            self._print_func(f" - Validation error: {validation_errors}")

        self._print_func()

        if abort_on_error:
            complain_and_abort("Errors encountered, exiting.")

        return False

    @staticmethod
    def _generate_significance_text(resource_modification: ResourceModification):
        if resource_modification.is_significant_change:
            return "\n   WARNING: This is a significant change which will reset/retrain the monitor."
        return ""

    @staticmethod
    def _echo_misconfigured_warnings(misconfigured_warnings: List[Dict[str, Any]]):
        for group in misconfigured_warnings:
            click.echo(f" - {group['title']}:")
            for item in group["items"]:
                click.echo(f"     - {item}")

    @manage_errors
    def compile(
        self, apply_namespace, abort_on_error=True
    ) -> Tuple[List[str], Dict[str, Box], Dict[str, Any]]:
        """
        Gather monitor configuration YAML files from project.
        This may also include dbt schema files with meta properties
        """
        self._init_local_config()
        self._print_func()
        self._print_func("Gathering monitor configuration files.")
        assert self.project_config is not None
        files = self._gather_yaml_files(self.project_config)

        if self.project_config.namespace:
            self._print_func(
                f"namespace: {self.project_config.namespace} found in {PROJECT_CONFIG_FILENAME} ignoring value passed with --namespace={apply_namespace}"
            )
            apply_namespace = self.project_config.namespace

        if self.dbt_manifest_path:
            self._init_dbt_config()

        errors_by_file = {}
        ns_mc_config_list_map: Dict[str, List[Box]] = {}

        found_config = False

        files = sorted(list(files))
        for file in files:
            with open(file, "r") as f:
                try:
                    yaml_as_dict = Box(yaml.safe_load(f))
                except Exception:
                    self._print_func(f" - Skipping {file}, not a valid YAML file.")
                    continue
                namespace = apply_namespace
                if has_namespace_property(yaml_as_dict):
                    # override apply_namespace with namespace in the yaml file.
                    namespace = yaml_as_dict.namespace

                if not namespace:
                    errors_by_file[file] = [
                        f"A default namespace need to be passed through command line --namespace or "
                        f"set default namespace in {PROJECT_CONFIG_FILENAME} or "
                        f"an override namespace needs to be set in file: {file}"
                    ]
                    continue
                if namespace not in ns_mc_config_list_map:
                    ns_mc_config_list_map[namespace] = []
                mc_config_list = ns_mc_config_list_map[namespace]

                # This file is a dbt schema file.
                # Let's attempt to parse out special embedded MC sections from meta:
                # See https://docs.getdbt.com/reference/resource-properties/meta
                if is_dbt_schema(yaml_as_dict):
                    for model in yaml_as_dict.models:
                        with contextlib.suppress(KeyError):
                            if model.meta:
                                mc_config = model.meta.montecarlo
                                errors = self._validate_mc_config(mc_config)
                                if self.dbt_manifest_path:
                                    mc_config = Box(self._resolve_templates(mc_config))
                                mc_config_list.append(mc_config)
                                found_config = True
                            else:
                                errors = ["meta tag should be removed if its empty"]
                            if errors:
                                errors_by_file[file] = errors
                            self._print_func(
                                f" - {file} - Embedded monitor configuration found."
                            )

                # This file has a root montecarlo property
                # This can be in a standalone file, and also a dbt schema file
                if has_montecarlo_property(yaml_as_dict):
                    mc_config = yaml_as_dict.montecarlo
                    errors = self._validate_mc_config(mc_config)
                    if self.dbt_manifest_path:
                        mc_config = Box(self._resolve_templates(mc_config))
                    if errors:
                        errors_by_file[file] = errors
                    mc_config_list.append(mc_config)
                    found_config = True
                    self._print_func(f" - {file} - Monitor configuration found.")

        # Exit if there are errors
        if errors_by_file:
            self._print_func()
            self._print_func("Configuration validation errors:")
            for file, errors in errors_by_file.items():
                self._print_func(f" - File: {file}")
                for error in errors:
                    self._print_func(f"    - {error}")
            self._print_func()
            if abort_on_error:
                complain_and_abort("Errors encountered, exiting.")

        if (not found_config) and abort_on_error:
            complain_and_abort(
                "Sorry, we didn't find any YAML files containing "
                "Monte Carlo configuration."
            )

        ns_compiled_mc_config: Dict[str, Box] = {}
        for namespace, mc_config_list in ns_mc_config_list_map.items():
            # Merge configs into a single config
            compiled_mc_config = self._merge_mc_configs(mc_config_list)

            # Write configs to target directory. Useful for debugging.
            with open(
                os.path.join(
                    self.target_dir, f"{namespace}_montecarlo_configuration.yml"
                ),
                "w",
            ) as f:
                f.write(compiled_mc_config.to_yaml())

            with open(
                os.path.join(
                    self.target_dir, f"{namespace}_montecarlo_configuration.json"
                ),
                "w",
            ) as f:
                json.dump(compiled_mc_config, f, indent=4)

            ns_compiled_mc_config[namespace] = compiled_mc_config

        return files, ns_compiled_mc_config, errors_by_file

    def _create_target_directory(self) -> str:
        target_dir = os.path.join(self.project_dir, TARGET_DIRECTORY_NAME)
        mkdirs(target_dir)
        return target_dir

    def _load_project_config(self) -> Optional[ProjectConfig]:
        contents = None
        try:
            with open(
                os.path.join(self.project_dir, PROJECT_CONFIG_FILENAME), "r"
            ) as f:
                contents = f.read()
        except FileNotFoundError:
            complain_and_abort(
                f"Not a Monte Carlo project. Must define {PROJECT_CONFIG_FILENAME} "
                "in current working directory."
            )

        assert contents is not None
        project_config_as_dict = yaml.safe_load(contents)
        try:
            s = ProjectConfig.schema()  # type: ignore
            return s.load(project_config_as_dict)
        except ValidationError as e:
            complain_and_abort(
                f"Encountered a validation problem in {PROJECT_CONFIG_FILENAME}: "
                f"{e.messages}"
            )

    def _merge_mc_configs(self, mc_config_list: List[Box]) -> Box:
        """
        Merge multiple mc configs into a single one.
        Nested dicts will be merged
        Nested lists will be concatenated
        """

        def merge_configs(x: Box, y: Box):
            # box_merge_lists="extend" does not recurse, so we need to process the
            # first config level ourselves to account for groups (like 'notifications')
            for (key, value) in y.items():
                if key in x:
                    if isinstance(value, dict):
                        x[key].merge_update(y[key], box_merge_lists="extend")
                    elif isinstance(value, list):
                        x[key].extend(y[key])
                else:
                    x[key] = value

            return x

        compiled_mc_config = reduce(merge_configs, mc_config_list)
        compiled_mc_config.pop(
            "box_merge_lists", None
        )  # Strangely enough, Box adds this add'l key. Remove it
        return compiled_mc_config

    def _validate_mc_config(self, mc_config_dict: Box) -> List[str]:
        """
        Perform simple validation.
        More comprehensive validation is performed server-side
        """
        expected_keys = [
            "field_health",
            "dimension_tracking",
            "json_schema",
            "custom_sql",
            "freshness",
            "volume",
        ]

        errors = []
        for key in expected_keys:
            with contextlib.suppress(KeyError):
                if key in mc_config_dict and not isinstance(mc_config_dict[key], list):
                    errors.append(f'"{key}" property should be a list.')
        return errors

    def _gather_yaml_files(self, project_config: ProjectConfig) -> Set[str]:
        files = set()
        for pattern in project_config.include_file_patterns:  # type: ignore
            for fn in glob.glob(
                os.path.join(self.project_dir, pattern), recursive=True
            ):
                files.add(fn)
        for pattern in project_config.exclude_file_patterns:  # type: ignore
            excluded = set(
                fnmatch.filter(files, os.path.join(self.project_dir, pattern))
            )
            files = files - excluded
        return files

    def _get_dbt_project_name(self) -> str:
        dbt_project = os.path.join(self.project_dir, "dbt_project.yml")
        try:
            with open(dbt_project, "r") as f:
                return yaml.safe_load(f)["name"]
        except Exception:
            complain_and_abort(f"Failed to read dbt project name from {dbt_project}.")

    def _load_dbt_manifest_nodes(self) -> dict:
        try:
            with open(self.dbt_manifest_path, "r") as f:
                return json.load(f)["nodes"]
        except Exception:
            complain_and_abort(
                f"Failed to load dbt manifest from {self.dbt_manifest_path}."
            )

    def _resolve_dbt_ref(self, match, project_delimiter: str = ":"):
        model = match.group(1)
        node = self.dbt_nodes.get(f"model.{self.dbt_project}.{model}")
        if not node:
            complain_and_abort(f"Could not resolve dbt ref for {model}.")

        database = node["database"]
        schema = node["schema"]
        name = node.get("alias") or node["name"]

        return f"{database}{project_delimiter}{schema}.{name}"

    def _resolve_sql_dbt_ref(self, match):
        # We use BQ `:` internally, but other warehouses use `.` so for SQL we need to
        # avoid using `:`. BQ supports `.` as well as `:` so we can just always use `.`.
        return self._resolve_dbt_ref(match, project_delimiter=".")

    def _resolve_templates(self, yaml_object, ref_resolver: Optional[Callable] = None):
        ref_resolver = ref_resolver or self._resolve_dbt_ref

        if isinstance(yaml_object, str):
            return self.DBT_REF_PATTERN.sub(ref_resolver, yaml_object)
        elif isinstance(yaml_object, list):
            return [self._resolve_templates(i, ref_resolver) for i in yaml_object]
        elif isinstance(yaml_object, dict):
            return {
                k: self._resolve_templates(
                    v,
                    self._resolve_sql_dbt_ref if k in self.SQL_KEYS else ref_resolver,
                )
                for k, v in yaml_object.items()
            }

        return yaml_object

    @manage_errors
    def delete(self, namespace: str, dry_run: bool = False, abort_on_error=True):
        """
        Delete all monitors in namespace
        """
        response = self.client.delete_config_template(namespace, dry_run=dry_run)

        dry_run_s = "[DRY RUN] " if dry_run else ""
        self._print_func(
            f"\n{dry_run_s}Deleted {response.num_deleted} "
            f"resources in namespace={namespace}\n"
        )

        return response

    @manage_errors
    def list_namespaces(self, limit=100):
        """
        Get all namespaces
        """
        query = Query()
        templates = query.get_monte_carlo_config_templates(first=limit + 1)
        assert templates.edges is not None
        templates.edges.node.__fields__("namespace", "last_update_time")
        response = self._pycarlo_client(query)
        namespaces = [
            (edge.node.namespace, edge.node.last_update_time)
            for edge in response.get_monte_carlo_config_templates.edges  # type: ignore
            if edge.node is not None
        ]

        more_ns_available = False
        if len(namespaces) > limit:
            namespaces = namespaces[:-1]
            more_ns_available = True
        self._print_func(
            tabulate(namespaces, headers=self.NS_HEADERS, tablefmt="fancy_grid")
        )

        if more_ns_available:
            self._print_func(self.MORE_NS_MESSAGE)

    @manage_errors
    def generate_from_dbt_tests(
        self,
        output_path: str,
        test_types: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
    ):
        """
        Create Monte Carlo monitor config YAML from dbt tests in dbt manifest.
        May specify a subset of dbt test types to import.
        """
        tests = (
            n
            for n in self._load_dbt_manifest_nodes().values()
            if n.get("resource_type") == "test"
        )

        if test_types:
            test_type_set = set(test_types)
            tests = (
                t
                for t in tests
                if t.get("test_metadata", {}).get("name", "SINGULAR") in test_type_set
                # Singular tests do not have metadata
            )

        common_args = {
            "schedule": {
                "type": "fixed",
                "start_time": datetime.now(timezone.utc).isoformat(),
                "interval_minutes": 720,
            },
            "comparisons": [
                {"type": "threshold", "operator": "GT", "threshold_value": 0}
            ],
        }

        if labels:
            common_args["labels"] = sorted(list(set(labels)))

        custom_sql_rules = [
            {
                "name": t["unique_id"],
                "description": t["unique_id"],
                "sql": t["compiled_code"].strip(),
                **common_args,
            }
            for t in tests
        ]

        self._print_func(f"Found {len(custom_sql_rules)} dbt tests.")

        Box({"montecarlo": {"custom_sql": custom_sql_rules}}).to_yaml(
            output_path, sort_keys=False, width=float("inf")
        )

        self._print_func(f"Wrote monitor config to {output_path}.")

    @staticmethod
    def _str_presenter(dumper, data):
        # Configures yaml for dumping multiline strings
        if data.count("\n") > 0:
            # It is a multiline string. Use a | style
            return dumper.represent_scalar(
                tag="tag:yaml.org,2002:str",
                value=MonteCarloConfigService._prepare_multiline_str(data),
                style="|",
            )

        return dumper.represent_scalar("tag:yaml.org,2002:str", data)

    @staticmethod
    def _prepare_multiline_str(data: str) -> str:
        """
        Prepare the data to format as multiline string in the YAML. It must meet
        some conditions to represent it as a block by yaml.dump,
        """

        # Trailing spaces are not permitted (check analyze_scalar in emitter.py,
        # and how allow_block is defined)
        result = "\n".join([line.rstrip() for line in data.splitlines()])
        if not result.endswith("\n"):
            # Add a line break at the end (to avoid using |- and use | in the yaml)
            result += "\n"

        return result



from typing import Optional

import click

from montecarlodata.collector.fields import (
    EXPECTED_TTC_RESPONSE_FIELD,
    EXPECTED_TTOC_RESPONSE_FIELD,
)
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.queries.collector import (
    TEST_TELNET_CONNECTION,
    TEST_TCP_OPEN_CONNECTION,
)
from montecarlodata.utils import GqlWrapper


class CollectorNetworkTestService:
    def __init__(self, config: Config, request_wrapper: Optional[GqlWrapper] = None):
        self._abort_on_error = True

        self._request_wrapper = request_wrapper or GqlWrapper(config)

    @manage_errors
    def echo_telnet_test(self, **kwargs):
        """
        Checks if telnet connection is usable and echos results
        """
        self._echo_network_test_validations(
            query=TEST_TELNET_CONNECTION,
            operation=EXPECTED_TTC_RESPONSE_FIELD,
            **kwargs
        )

    @manage_errors
    def echo_tcp_open_test(self, **kwargs):
        """
        Checks if a destination exists and accepts requests and echos results
        """
        self._echo_network_test_validations(
            query=TEST_TCP_OPEN_CONNECTION,
            operation=EXPECTED_TTOC_RESPONSE_FIELD,
            **kwargs
        )

    def _echo_network_test_validations(self, query: str, operation: str, **kwargs):
        """
        Do request and echos a validation response
        """
        response = self._request_wrapper.make_request_v2(
            query=query, operation=operation, variables=kwargs
        )
        click.echo(response.data.validations[0].message)



import traceback
from functools import wraps
from typing import Any, Optional, List

import click

import montecarlodata.settings as settings
from montecarlodata.common.data import MonolithResponse


def manage_errors(func):
    """
    Convenience decorator to abort on any errors after logging based on verbosity settings

    Requires an `_abort_on_error` field to be set in the instance
    """

    @wraps(func)
    def _impl(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except click.Abort:
            raise click.Abort()  # re-raise to prevent swallowing
        except Exception as error:
            if hasattr(self, "_disable_handle_errors") and self._disable_handle_errors:
                raise
            echo_error(error, [traceback.format_exc()])

        if hasattr(self, "_abort_on_error") and self._abort_on_error:
            raise click.Abort()

    return _impl


def echo_error(message: Any, errors: Optional[List[Any]] = None) -> None:
    """
    Convenience utility to echo any error in verbose and quiet mode
    """
    click.echo(f"Error - {message}", err=True)
    if settings.MCD_VERBOSE_ERRORS:
        for error in errors or []:
            click.echo(error, err=True)


def complain_and_abort(message: str) -> None:
    """
    Convenience utility to echo message and exit
    """
    click.echo(f"Error - {message}", err=True)
    raise click.Abort()


def abort_on_gql_errors(response: MonolithResponse):
    """
    Convenience utility to echo any gql errors and exit
    """
    if response.errors:
        for error in response.errors:
            echo_error(error.get("message"))
        raise click.Abort()


def prompt_connection(message: str, skip_prompt: Optional[bool] = False) -> None:
    """
    Prompt message and abort if not confirmed
    """
    if not skip_prompt:
        click.confirm(message, abort=True)



import configparser
from collections import OrderedDict
from unittest import TestCase
from unittest.mock import Mock, patch, mock_open, call

from montecarlodata.config import Config, ConfigManager

_SAMPLE_PROFILE_NAME = "foo"
_SAMPLE_BASE_PATH = "bar"
_SAMPLE_FILE = f"{_SAMPLE_BASE_PATH}/profiles.ini"
_SAMPLE_OPTIONS = OrderedDict(
    {
        "mcd_id": "1234",
        "mcd_token": "a-test-token-with-fifty-six-characters-for-verifications",
        "mcd_api_endpoint": "https://api.getmontecarlo.com/graphql",
        "aws_profile": "foo",
        "aws_region": "home",
    }
)


class ConfigTest(TestCase):
    def setUp(self) -> None:
        self._parser_mock = Mock()

        self._service = ConfigManager(
            profile_name=_SAMPLE_PROFILE_NAME,
            base_path=_SAMPLE_BASE_PATH,
            config_parser=self._parser_mock,
        )

    def test_setup(self):
        self._parser_mock.read.assert_called_once_with(_SAMPLE_FILE)

    @patch("montecarlodata.config.mkdirs")
    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_write(self, mock_file, create_directory_mock):
        self._parser_mock.sections.return_value = []
        self._service.write(**_SAMPLE_OPTIONS)

        self._parser_mock.add_section.assert_called_once_with(_SAMPLE_PROFILE_NAME)

        self._parser_mock.set.assert_has_calls(
            [call(_SAMPLE_PROFILE_NAME, k, v) for k, v in _SAMPLE_OPTIONS.items()]
        )

        mock_file.assert_called_once_with(_SAMPLE_FILE, "w")
        create_directory_mock.assert_called_once_with(_SAMPLE_BASE_PATH)

    @patch("montecarlodata.errors.echo_error")
    @patch("montecarlodata.config.mkdirs")
    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_write_invalid_token(self, mock_file, create_directory_mock, mock_echo):
        self._parser_mock.sections.return_value = []
        INVALID_SHORT_TOKEN = "invalid-short-token"

        self._service.write(**{**_SAMPLE_OPTIONS, "mcd_token": INVALID_SHORT_TOKEN})
        mock_echo.assert_called_once()

        self._parser_mock.add_section.assert_called_once_with(_SAMPLE_PROFILE_NAME)
        self._parser_mock.set.called_once_with(
            _SAMPLE_PROFILE_NAME, "mcd_token", INVALID_SHORT_TOKEN
        )

        mock_file.assert_not_called()
        create_directory_mock.assert_not_called()

    def test_read(self):
        self._parser_mock.get.side_effect = _SAMPLE_OPTIONS.values()
        self.assertEqual(self._service.read(), Config(**_SAMPLE_OPTIONS))

    def test_read_with_no_section(self):
        self._parser_mock.get.side_effect = configparser.NoSectionError(
            _SAMPLE_PROFILE_NAME
        )
        self.assertIsNone(self._service.read())



from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.transactional.transactional_db import (
    TransactionalOnboardingService,
)
from montecarlodata.queries.onboarding import (
    TEST_DATABASE_CRED_MUTATION,
)
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.test_base_onboarding import _SAMPLE_BASE_OPTIONS
from tests.test_common_user import _SAMPLE_CONFIG


class TransactionalOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = TransactionalOnboardingService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    @patch.object(TransactionalOnboardingService, "onboard")
    def test_generic_transactional_db_flow(self, onboard_mock):
        expected_options = {
            **{
                "connectionType": "transactional-db",
                "warehouseType": "transactional-db",
            },
            **_SAMPLE_BASE_OPTIONS,
        }

        self._service.onboard_transactional_db(**_SAMPLE_BASE_OPTIONS)
        onboard_mock.assert_called_once_with(
            validation_query=TEST_DATABASE_CRED_MUTATION,
            validation_response="testDatabaseCredentials",
            connection_type="transactional-db",
            **expected_options
        )



"""
Common commands
"""
import click

from montecarlodata.tools import convert_uuid_callback

RESOURCE_VERBIAGE = (
    "This can be helpful if the resource and collector are in different accounts"
)

DISAMBIGUATE_DC_OPTIONS = [
    click.option(
        "--collector-id",
        "dc_id",
        required=False,
        type=click.UUID,
        callback=convert_uuid_callback,
        help="ID for the data collector. To disambiguate accounts with multiple collectors.",
    ),
]

DC_RESOURCE_OPTIONS = [
    click.option(
        "--resource-aws-region",
        required=False,
        help="Override the AWS region where the resource is located. "
        "Defaults to the region where the collector is hosted.",
    ),
    click.option(
        "--resource-aws-profile",
        required=False,
        help=f"Override the AWS profile use by the CLI for the resource. {RESOURCE_VERBIAGE}.",
    ),
]



from dataclasses import dataclass
from typing import Optional, Dict, List, Union
from uuid import UUID

from box import Box

from montecarlodata.common.common import boxify
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort
from montecarlodata.queries.user import GET_USER_QUERY
from montecarlodata.utils import GqlWrapper


@dataclass
class User:
    first_name: Optional[str] = None
    last_name: Optional[str] = None


class UserService:
    def __init__(self, config: Config, request_wrapper: Optional[GqlWrapper] = None):
        self._config = config
        self._request_wrapper = request_wrapper or GqlWrapper(config)

        self._user = self._request_wrapper.make_request(GET_USER_QUERY)  # get user info

    @property
    def user(self) -> User:
        """
        Returns basic user properties
        """
        return User(
            first_name=self._user["getUser"].get("firstName"),
            last_name=self._user["getUser"].get("lastName"),
        )

    @property
    def account(self) -> Dict:
        """
        Get account details
        """
        return self._user["getUser"]["account"]

    @property
    def collectors(self) -> Dict:
        """
        Get collectors in the account
        """
        return self.account.get("dataCollectors", [{}])

    @property
    def active_collector(self) -> Dict:
        """
        Get active collector from collectors.

        Errors out on accounts with > 1 collector. Legacy. Do not use. See `get_collector`
        """
        return self.collectors[self._get_active_collector()]

    @property
    def resource_identifiers(self) -> Dict:
        """
        Returns a commutative mapping of resource uuid and name.
        e.g. user_service.resource_identifiers().get("resource_name") == "resource_uuid"
         and user_service.resource_identifiers().get("resource_uuid") == "resource_name"

        Currently only supports warehouse resources.
        """
        resource_identifiers = {}
        for resource in self.warehouses:
            resource_identifiers[resource["uuid"]] = resource["name"]
            resource_identifiers[resource["name"]] = resource["uuid"]
        return resource_identifiers

    @property
    def warehouses(self) -> List[Dict]:
        """
        Get warehouses attached to the account
        """
        return self.account.get("warehouses")

    @property
    def bi_containers(self) -> List[Dict]:
        """
        Get bi attached to the account
        """
        return self.account.get("bi")

    @property
    def etl_containers(self) -> List[Dict]:
        """
        Get etl attached to the account
        """
        return self.account.get("etlContainers")

    @property
    def tableau_accounts(self) -> Dict:
        """
        Get tableau connections attached to the account
        """
        return self.account.get("tableauAccounts")

    @property
    def active_collection_regions(self) -> List[str]:
        """
        Get a list of active collection regions
        """
        return self.account.get("activeCollectionRegions")

    @boxify()
    def get_collector(self, dc_id: Optional[str] = None) -> Box:
        """
        Get a specific collector

        This is only necessary for client only ops on a specific collector. APIs handle this disambiguation already.
        """
        num_of_collectors = len(self.collectors)
        if num_of_collectors == 0:
            complain_and_abort("No collector found.")

        if dc_id:
            for collector in self.collectors:
                if dc_id == collector["uuid"]:
                    return collector
            complain_and_abort(f"Collector with ID '{dc_id}' not found.")
        if num_of_collectors > 1:
            complain_and_abort(
                "Multiple data collectors found. Please specify a collector."
            )
        return self.collectors[0]

    def _get_active_collector(self) -> Optional[int]:
        """
        Get active collector - currently only one active collector per account is supported. Abort if None are found
        """
        if len(self.collectors) > 1:
            complain_and_abort(
                "This option is only supported in accounts with one collector."
            )

        for idx, collector in enumerate(self.collectors):
            if collector.get("active"):
                return idx
        complain_and_abort("No active collector found")

    def get_warehouse_for_connection(
        self, connection_id: Union[str, UUID]
    ) -> Optional[Box]:
        """
        Get warehouse for the given connection id
        """
        for warehouse in self.warehouses:
            for connection in warehouse.get("connections", []):
                if str(connection_id) == connection.get("uuid"):
                    return Box(warehouse)



from unittest import TestCase
from unittest.mock import Mock, patch
from uuid import uuid4

import click
from box import Box
from pycarlo.features.dbt import DbtImporter
from pycarlo.features.exceptions import MultipleResourcesFoundException

from montecarlodata.dataimport.dbt import DbtImportService


class DbtImportServiceTests(TestCase):

    project_name = "default-project"
    job_name = "default-job"
    manifest_path = "./manifest.json"
    run_results_path = "./run_results.json"
    logs_path = "./logs.txt"

    def setUp(self):
        self.mock_config = Mock()
        self.mock_mc_client = Mock()
        self.mock_user_service = Mock()
        self.mock_pii_service = Mock()
        self.mock_pii_service.get_pii_filters_config.return_value = None

        self.service = DbtImportService(
            config=self.mock_config,
            mc_client=self.mock_mc_client,
            user_service=self.mock_user_service,
            pii_service=self.mock_pii_service,
        )

    @patch.object(DbtImporter, "import_run")
    def test_import_run_with_no_connection_id(self, mock_import_run: Mock):
        # given
        connection_id = None

        # when
        self.service.import_run(
            project_name=self.project_name,
            job_name=self.job_name,
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            connection_id=connection_id,
        )

        # then
        self.mock_user_service.get_warehouse_for_connection.assert_not_called()
        mock_import_run.assert_called_once_with(
            project_name=self.project_name,
            job_name=self.job_name,
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=None,
        )

    @patch.object(click, "echo")
    @patch.object(DbtImporter, "import_run")
    def test_import_run_with_no_connection_id_and_multiple_warehouses(
        self, mock_import_run: Mock, mock_echo: Mock
    ):
        # given
        connection_id = None
        mock_import_run.side_effect = MultipleResourcesFoundException()

        # when
        with self.assertRaises(click.Abort):
            self.service.import_run(
                project_name=self.project_name,
                job_name=self.job_name,
                manifest_path=self.manifest_path,
                run_results_path=self.run_results_path,
                logs_path=self.logs_path,
                connection_id=connection_id,
            )

        # then
        self.mock_user_service.get_warehouse_for_connection.assert_not_called()
        mock_import_run.assert_called_once_with(
            project_name=self.project_name,
            job_name=self.job_name,
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=None,
        )
        mock_echo.assert_called_once_with(
            "Error - Multiple resources found, please specify a connection id.",
            err=True,
        )

    @patch.object(click, "echo")
    @patch.object(DbtImporter, "import_run")
    def test_import_run_with_invalid_connection_id(
        self, mock_import_run: Mock, mock_echo: Mock
    ):
        # given
        connection_id = uuid4()
        self.mock_user_service.get_warehouse_for_connection.return_value = None

        # when
        with self.assertRaises(click.Abort):
            self.service.import_run(
                project_name=self.project_name,
                job_name=self.job_name,
                manifest_path=self.manifest_path,
                run_results_path=self.run_results_path,
                logs_path=self.logs_path,
                connection_id=connection_id,
            )

        # then
        self.mock_user_service.get_warehouse_for_connection.assert_called_once_with(
            connection_id
        )
        mock_import_run.assert_not_called()
        mock_echo.assert_called_once_with(
            f"Error - Could not find a connection with id: {connection_id}", err=True
        )

    @patch.object(DbtImporter, "import_run")
    def test_import_run_with_valid_connection_id(self, mock_import_run: Mock):
        # given
        connection_id = uuid4()
        resource_id = str(uuid4())
        self.mock_user_service.get_warehouse_for_connection.return_value = Box(
            {"uuid": resource_id}
        )

        # when
        self.service.import_run(
            project_name=self.project_name,
            job_name=self.job_name,
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            connection_id=connection_id,
        )

        # then
        self.mock_user_service.get_warehouse_for_connection.assert_called_once_with(
            connection_id
        )
        mock_import_run.assert_called_once_with(
            project_name=self.project_name,
            job_name=self.job_name,
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=resource_id,
        )



import csv
import sys
from dataclasses import dataclass
from typing import Optional, List, Dict

import click
from dataclasses_json import dataclass_json, LetterCase
from pycarlo.core import Client, Mutation, Query
from tabulate import tabulate

from montecarlodata.config import Config
from montecarlodata.common.common import ConditionalDictionary
from montecarlodata.common.user import UserService
from montecarlodata.errors import manage_errors, complain_and_abort

DEFAULT_COLLECTION_BLOCK_LIST_PAGE_SIZE = 100


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class CollectionBlock:
    resource_id: str
    project: str
    dataset: Optional[str] = None


class ManagementService:
    _LIST_PII_PREFERENCES_HEADERS = ["Status", "Fail Mode"]
    _COLLECTION_BLOCK_LIST_HEADERS = ["Resource", "Project", "Dataset"]
    _mutation_add_to_block_list = """
        mutation addToCollectionBlockList($collectionBlocks: [CollectionBlockInput]!) {
            addToCollectionBlockList(collectionBlocks: $collectionBlocks) {
                success
            }
        }
    """
    _mutation_remove_from_block_list = """
        mutation removeFromCollectionBlockList($collectionBlocks: [CollectionBlockInput]!) {
            removeFromCollectionBlockList(collectionBlocks: $collectionBlocks) {
                success
            }
        }
    """

    def __init__(
        self,
        config: Config,
        mc_client: Client,
        user_service: Optional[UserService] = None,
    ):
        self._abort_on_error = True
        self._mc_client = mc_client
        self._user_service = user_service or UserService(config=config)

    @manage_errors
    def get_pii_preferences(
        self,
        headers: Optional[str] = "firstrow",
        table_format: Optional[str] = "fancy_grid",
    ) -> None:
        table = [self._LIST_PII_PREFERENCES_HEADERS]

        query = Query()
        query.get_pii_filtering_preferences()
        preferences = self._mc_client(query).get_pii_filtering_preferences
        table.append(
            [
                "Enabled" if preferences.enabled else "Disabled",
                preferences.fail_mode,
            ]
        )
        click.echo(tabulate(table, headers=headers, tablefmt=table_format))

    @manage_errors
    def set_pii_filtering(
        self,
        enabled: Optional[bool] = None,
        fail_mode: Optional[str] = None,
    ) -> None:
        variables = ConditionalDictionary(lambda x: x is not None)
        variables.update({"enabled": enabled, "fail_mode": fail_mode})

        mutation = Mutation()
        mutation.update_pii_filtering_preferences(**variables)
        result = self._mc_client(mutation).update_pii_filtering_preferences
        if result.success:
            click.echo("PII filtering preferences have been updated!")

    @manage_errors
    def get_collection_block_list(
        self,
        resource_name: Optional[str] = None,
        limit: Optional[int] = DEFAULT_COLLECTION_BLOCK_LIST_PAGE_SIZE,
        headers: Optional[str] = "firstrow",
        table_format: Optional[str] = "fancy_grid",
    ) -> None:
        table = [self._COLLECTION_BLOCK_LIST_HEADERS]

        variables = {"first": limit}
        if resource_name:
            resource_id = self._user_service.resource_identifiers.get(resource_name)
            if not resource_id:
                sys.exit(f"No existing resources with name '{resource_name}'.")
            variables["resource_id"] = resource_id

        while True:  # loop through each page of results
            query = Query()
            op = query.get_collection_block_list(**variables)
            op.edges.node.__fields__("resource_id", "project", "dataset")
            op.page_info()

            result = self._mc_client(query)
            collection_block_connection = result.get_collection_block_list
            for edge in collection_block_connection.edges:
                table.append(
                    [
                        self._user_service.resource_identifiers[edge.node.resource_id],
                        edge.node.project,
                        edge.node.dataset,
                    ]
                )
            if not collection_block_connection.page_info.has_next_page:
                break
            variables["after"] = collection_block_connection.page_info.end_cursor

        click.echo(tabulate(table, headers=headers, tablefmt=table_format))

    def update_collection_block_list(
        self,
        adding: bool,
        filename: Optional[str] = None,
        resource_name: Optional[str] = None,
        project: Optional[str] = None,
        dataset: Optional[str] = None,
    ):
        mutation = (
            self._mutation_add_to_block_list
            if adding
            else self._mutation_remove_from_block_list
        )

        resource_id = None
        if resource_name:
            resource_id = self._user_service.resource_identifiers.get(resource_name)
            if not resource_id:
                complain_and_abort(
                    f"No existing resources with name '{resource_name}'."
                )

        collection_blocks_input = (
            self.parse_collection_blocks_file(filename)
            if filename
            else [
                {
                    "resourceId": resource_id,
                    "project": project,
                    "dataset": dataset,
                },
            ]
        )
        operation = self._mc_client(
            mutation, variables={"collectionBlocks": collection_blocks_input}
        )
        result = (
            operation.add_to_collection_block_list
            if adding
            else operation.remove_from_collection_block_list
        )
        if result.success:
            click.echo("Collection block list has been updated!")

    def parse_collection_blocks_file(self, filename: str) -> List[Dict]:
        block_list_file = csv.DictReader(open(filename))
        collection_blocks = []
        for row in block_list_file:
            resource_name = row["resource_name"]
            resource_id = self._user_service.resource_identifiers.get(resource_name)
            if not resource_id:
                complain_and_abort(
                    f"No existing resources with name '{resource_name}'."
                )

            collection_blocks.append(
                CollectionBlock(
                    resource_id=resource_id,
                    project=row["project"],
                    dataset=row.get("dataset"),
                ).to_dict()
            )
        return collection_blocks



CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC = """
mutation createOrUpdateMonteCarloConfigTemplateAsync(
    $namespace: String!,
    $configTemplateJson: String!,
    $dryRun: Boolean,
    $misconfiguredAsWarning: Boolean,
    $resource: String
) {
    createOrUpdateMonteCarloConfigTemplateAsync(
        configTemplateJson: $configTemplateJson,
        namespace: $namespace,
        dryRun: $dryRun,
        misconfiguredAsWarning: $misconfiguredAsWarning,
        resource: $resource
    ) {
        response {
            updateUuid
            errorsAsJson
        }
    }
}
"""

GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE = """
query getMonteCarloConfigTemplateUpdateState($updateUuid: UUID!) {
    getMonteCarloConfigTemplateUpdateState(updateUuid: $updateUuid) {
        state
        resourceModifications {
            type
            description
            resourceAsJson
        }
        changesApplied
        errorsAsJson
        warningsAsJson
    }
}
"""

DELETE_MONTE_CARLO_CONFIG_TEMPLATE = """
mutation deleteMonteCarloConfigTemplate($namespace: String!, $dryRun: Boolean) {
  deleteMonteCarloConfigTemplate(
    namespace: $namespace,
    dryRun: $dryRun
  ) {
    response {
      changesApplied
      numDeleted
    }
  }
}
"""



from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.integrations.onboarding.base import BaseOnboardingService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_DBT_CLOUD_RESPONSE_FIELD,
    EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
    DBT_CLOUD_CONNECTION_TYPE,
    DBT_CLOUD_WEBHOOK_CONNECTION_TYPE,
)
from montecarlodata.queries.onboarding import (
    TEST_DBT_CLOUD_CRED_MUTATION,
    ADD_CONNECTION_MUTATION,
)


class DbtCloudOnboardingService(BaseOnboardingService):
    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    @manage_errors
    def onboard_dbt_cloud(self, **kwargs) -> None:
        connection_type = DBT_CLOUD_CONNECTION_TYPE

        hmac_secret = kwargs.pop("webhook_hmac_secret", None)
        webhook_id = kwargs.pop("webhook_id", None)
        if hmac_secret is not None:
            connection_type = DBT_CLOUD_WEBHOOK_CONNECTION_TYPE
            kwargs["dbt_cloud_webhook_hmac_secret"] = hmac_secret
        if webhook_id is not None:
            kwargs["dbt_cloud_webhook_id"] = webhook_id

        self.onboard(
            validation_query=TEST_DBT_CLOUD_CRED_MUTATION,
            validation_response=EXPECTED_DBT_CLOUD_RESPONSE_FIELD,
            connection_query=ADD_CONNECTION_MUTATION,
            connection_response=EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
            connection_type=connection_type,
            **kwargs
        )



import json
import uuid
from typing import Dict
from unittest import TestCase
from unittest.mock import Mock, patch

from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD,
    EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD,
    OPERATION_ERROR_VERBIAGE,
)
from montecarlodata.integrations.onboarding.operations.connection_ops import (
    ConnectionOperationsService,
)
from montecarlodata.queries.onboarding import (
    UPDATE_CREDENTIALS_MUTATION,
    REMOVE_CONNECTION_MUTATION,
    TEST_EXISTING_CONNECTION_QUERY,
)
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_UUID = uuid.uuid4()


class ConnectionOperationsTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = ConnectionOperationsService(
            _SAMPLE_CONFIG,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    def test_update_credentials(self):
        changes = {"host": "foo"}
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"success": True}
        )
        func = capture_function(
            self._service.update_credentials,
            dict(connection_id=_SAMPLE_UUID, changes=changes),
        )

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=UPDATE_CREDENTIALS_MUTATION,
            operation=EXPECTED_UPDATE_CREDENTIALS_RESPONSE_FIELD,
            variables=dict(
                connection_id=str(_SAMPLE_UUID),
                changes=json.dumps(changes),
                should_validate=True,
                should_replace=False,
            ),
        )
        self.assertEqual(
            func.std_out.getvalue().strip(), f"Success! Updated '{_SAMPLE_UUID}'."
        )

    def test_remove_connection(self):
        self._test_remove_connection(
            params=dict(connection_id=_SAMPLE_UUID, no_prompt=True)
        )

    @patch(
        "montecarlodata.integrations.onboarding.operations.connection_ops.click.confirm"
    )
    def test_remove_connection_with_prompt(self, confirm):
        confirm.return_value = True
        self._test_remove_connection(
            params=dict(connection_id=_SAMPLE_UUID, no_prompt=False)
        )

    def _test_remove_connection(self, params: Dict):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"success": True}
        )
        func = capture_function(self._service.remove_connection, params)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=REMOVE_CONNECTION_MUTATION,
            operation=EXPECTED_REMOVE_CONNECTION_RESPONSE_FIELD,
            variables=dict(connection_id=str(_SAMPLE_UUID)),
        )
        self.assertEqual(
            func.std_out.getvalue().strip(), f"Success! Removed '{_SAMPLE_UUID}'."
        )

    def test_echo_operation_status_with_errors(self):
        func = capture_function(
            function=self._service.echo_operation_status,
            params=dict(
                response=MonolithResponse(data={"success": False}), operation="foo"
            ),
        )  # Note - other tests capture the success state

        self.assertEqual(func.std_out.getvalue().strip(), OPERATION_ERROR_VERBIAGE)

    @patch("montecarlodata.integrations.onboarding.operations.connection_ops.click")
    def test_echo_test_existing(self, click_mock):
        connection_id = uuid.uuid4()
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"success": True}
        )
        self._service.echo_test_existing(connection_id=connection_id)
        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=TEST_EXISTING_CONNECTION_QUERY,
            operation="testExistingConnection",
            variables={"connection_id": connection_id},
        )
        click_mock.echo.assert_called_once_with('{\n    "success": true\n}')



import click
import click_config_file

import montecarlodata.settings as settings
from montecarlodata.collector.network_tests import CollectorNetworkTestService
from montecarlodata.collector.fields import DEFAULT_COLLECTION_REGION
from montecarlodata.collector.management import CollectorManagementService
from montecarlodata.common import create_mc_client
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS
from montecarlodata.tools import (
    add_common_options,
    convert_uuid_callback,
    validate_json_callback,
)

# Shared command verbiage
PROFILE_VERBIAGE = "If not specified, the one in the Monte Carlo CLI profile is used"
ADD_DC_VERBIAGE = (
    "Prompts to opens browser to CF console. If declined (or skipped) can use deploy, get-template "
    "or open-link with the generated ID."
)

# Options shared across commands
REGION_OPTIONS = [
    *DISAMBIGUATE_DC_OPTIONS,
    click.option(
        "--aws-region",
        required=False,
        default=DEFAULT_COLLECTION_REGION,
        show_default=True,
        help="AWS region where the collector is deployed or intended to be deployed.",
    ),
]
DEPLOYMENT_OPTIONS = [
    click.option(
        "--aws-profile", required=False, help=f"AWS profile. {PROFILE_VERBIAGE}."
    ),
    click.option(
        "--params",
        required=False,
        default=None,
        callback=validate_json_callback,
        help="""
                  Parameters key,value pairs as JSON. If a key is not specified the existing (or default) value is used.
                  \b
                  \n
                  E.g. --params '{"CreateEventInfra":"True"}'
                  """,
    ),  # \b disables wrapping
    *REGION_OPTIONS,
]

NETWORK_TEST_OPTIONS = [
    *DISAMBIGUATE_DC_OPTIONS,
    click.option("--host", required=True, help="Host to check."),
    click.option("--port", required=True, type=click.INT, help="Port to check."),
    click.option(
        "--timeout",
        required=False,
        default=5,
        type=click.INT,
        show_default=True,
        help="Timeout in seconds.",
    ),
]


@click.group(help="Manage a data collector.")
def collectors():
    """
    Group for any collector related subcommands
    """
    pass


@collectors.command(help="List all collector records.", name="list")
@click.pass_obj
@click.option(
    "--active-only",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Only list active collectors.",
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def list_collectors(ctx, active_only):
    CollectorManagementService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        aws_wrapper=None,
    ).echo_collectors(active_only=active_only)


@collectors.command(
    help=f"Add a collector record to the account. {ADD_DC_VERBIAGE}", name="add"
)
@click.pass_obj
@click.option(
    "--no-prompt",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Skip prompt for launching browser (Auto no).",
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def add_collector(ctx, no_prompt):
    CollectorManagementService(
        config=ctx["config"], mc_client=create_mc_client(ctx)
    ).add_collector(no_prompt=no_prompt)


@collectors.command(
    help="Get link to the latest template. For initial deployment or manually upgrading."
)
@click.pass_obj
@add_common_options(REGION_OPTIONS)
@click.option(
    "--update-infra/--no-update-infra",
    default=False,
    show_default=True,
    help="Update the collector infrastructure. Otherwise, only the lambda code will be updated.",
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def get_template(ctx, aws_region, **kwargs):
    """
    Get the latest template for this account
    """
    CollectorManagementService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        aws_region_override=aws_region,
    ).echo_template(**kwargs)


@collectors.command(
    help="Opens browser to CF console with a quick create link. For initial deployment."
)
@click.pass_obj
@click.option(
    "--dry",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Echos quick create link.",
)
@add_common_options(REGION_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def open_link(ctx, aws_region, **kwargs):
    """
    Open browser with a quick create link for deploying a data collector
    """
    CollectorManagementService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        aws_region_override=aws_region,
    ).launch_quick_create_link(**kwargs)


@collectors.command(help="Deploy a data collector stack.")
@click.pass_obj
@add_common_options(DEPLOYMENT_OPTIONS)
@click.option(
    "--stack-name",
    required=True,
    help="The name that is associated with the CloudFormation stack. "
    "Must be unique in the region.",
)
@click.option(
    "--enable-termination-protection/--no-enable-termination-protection",
    "termination_protection",
    default=False,
    show_default=True,
    help="Enable termination protection for this stack.",
)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def deploy(ctx, aws_profile, aws_region, params, **kwargs):
    """
    Deploy a collector for this account
    """
    CollectorManagementService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        aws_profile_override=aws_profile,
        aws_region_override=aws_region,
    ).deploy_template(new_params=params, **kwargs)


@collectors.command(help="Upgrade to the latest version.")
@click.pass_obj
@click.option(
    "--update-infra/--no-update-infra",
    default=False,
    show_default=True,
    help="Update the collector infrastructure. Otherwise, only the lambda code will be updated.",
)
@add_common_options(DEPLOYMENT_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def upgrade(ctx, aws_profile, aws_region, params, update_infra, **kwargs):
    """
    Upgrade the collector for this account
    """
    CollectorManagementService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        aws_profile_override=aws_profile,
        aws_region_override=aws_region,
    ).upgrade_template(update_infra=update_infra, new_params=params, **kwargs)


@collectors.command(help="Checks if telnet connection is usable from the collector.")
@click.pass_obj
@add_common_options(NETWORK_TEST_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def test_telnet(ctx, **kwargs):
    """
    Network debugging utility to test telnet via the collector
    """
    CollectorNetworkTestService(config=ctx["config"]).echo_telnet_test(**kwargs)


@collectors.command(
    help="Tests if a destination exists and accepts requests. "
    "Opens a TCP Socket to a specific port from the collector."
)
@click.pass_obj
@add_common_options(NETWORK_TEST_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def test_tcp_open(ctx, **kwargs):
    """
    Network debugging utility to test TCP open via the collector
    """
    CollectorNetworkTestService(config=ctx["config"]).echo_tcp_open_test(**kwargs)


@collectors.command(help="Validate whether a collector is operational.")
@click.pass_obj
@click.option(
    "--trace-id",
    required=False,
    type=click.UUID,
    callback=convert_uuid_callback,
    help=(
        "Optional custom UUID for tracking and correlating the ping response. "
        "By default we generate a random UUID to be sent with the ping request "
        "and validate that the ping response contains the same exact value."
    ),
)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def ping(ctx, **kwargs):
    CollectorManagementService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
    ).ping_dc(**kwargs)



import click
import click_config_file

from montecarlodata import settings
from montecarlodata.common import create_mc_client
from montecarlodata.iac.mc_config_service import MonteCarloConfigService
from montecarlodata.monitors.monitor_service import MonitorService


@click.group(help="Manage monitors.")
def monitors():
    """
    Group for any monitor related subcommands
    """
    pass


@monitors.command(help="Compile monitor configuration.")
@click.option(
    "--project-dir",
    required=False,
    type=click.Path(exists=True),
    help="Base directory of MC project (where montecarlo.yml is located). "
    "By default, this is set to the current working directory",
)
@click.option(
    "--namespace",
    required=False,
    help="Namespace of monitors configuration. This value will be ignored if we find a"
    " namespace in montecarlo.yml",
)
@click.option(
    "--dbt-manifest",
    required=False,
    type=click.Path(exists=True),
    help="Path to dbt manifest used for resolving dbt ref() in monitor config",
)
@click.pass_obj
def compile(ctx, project_dir, namespace, dbt_manifest):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        project_dir=project_dir,
        dbt_manifest_path=dbt_manifest,
    ).compile(namespace)


@monitors.command(help="Compile and apply monitor configuration.")
@click.option(
    "--project-dir",
    required=False,
    type=click.Path(exists=True),
    help="Base directory of MC project (where montecarlo.yml is located). "
    "By default, this is set to the current working directory",
)
@click.option(
    "--namespace",
    required=False,
    help="Namespace of monitors configuration. This value will be ignored if we find a"
    " namespace in montecarlo.yml",
)
@click.option(
    "--dry-run",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Dry run (just shows planned changes but doesn't apply them.)",
)
@click.option(
    "--dbt-manifest",
    required=False,
    type=click.Path(exists=True),
    help="Path to dbt manifest used for resolving dbt ref() in monitor config",
)
@click.option(
    "--auto-yes",
    is_flag=True,
    help="Skip any interactive approval.",
    default=False,
    show_default=True,
)
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def apply(ctx, project_dir, namespace, dry_run, dbt_manifest, auto_yes):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        project_dir=project_dir,
        dbt_manifest_path=dbt_manifest,
    ).apply(namespace, dry_run=dry_run, skip_confirmation=auto_yes)


@monitors.command(help="Delete monitor configuration.")
@click.option(
    "--project-dir",
    required=False,
    type=click.Path(exists=True),
    help="Base directory of MC project (where montecarlo.yml is located). "
    "By default, this is set to the current working directory",
)
@click.option("--namespace", required=True, help="Namespace of monitors configuration.")
@click.option(
    "--dry-run",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Dry run (just shows planned changes but doesn't apply them.",
)
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def delete(ctx, project_dir, namespace, dry_run):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        project_dir=project_dir,
    ).delete(namespace, dry_run=dry_run)


@monitors.command(name="list", help="List monitors ordered by update recency.")
@click.option(
    "--limit",
    required=False,
    default=100,
    type=click.INT,
    show_default=True,
    help="Max number of monitors to list.",
)
@click.option(
    "--monitor-type",
    required=False,
    type=click.Choice(MonitorService.MONITOR_TYPES, case_sensitive=False),
    help="List monitors with monitor_type",
)
@click.option(
    "--namespace", required=False, help="List only monitors in this namespace"
)
@click.pass_obj
def list_monitors(ctx, monitor_type, limit, namespace):
    monitor_types = None
    namespaces = [namespace] if namespace else None
    if monitor_type:
        if monitor_type == MonitorService.PSEUDO_MONITOR_TYPE_CB_COMPATIBLE:
            monitor_type = MonitorService.MONITOR_TYPE_CUSTOM_SQL
        monitor_types = [monitor_type]
    MonitorService(client=create_mc_client(ctx)).list_monitors(
        limit, namespaces, monitor_types
    )


@monitors.command(name="namespaces", help="List all namespaces.")
@click.option(
    "--limit",
    required=False,
    default=100,
    type=click.INT,
    show_default=True,
    help="Max number of namespaces to list.",
)
@click.pass_obj
def list_namespaces(ctx, limit):
    MonteCarloConfigService(
        config=ctx["config"], pycarlo_client=create_mc_client(ctx)
    ).list_namespaces(limit)


@monitors.command(help="Generate MC monitor config YAML from dbt tests.")
@click.option(
    "--output-path",
    required=True,
    type=click.Path(),
    help="Path for MC monitor config output YAML",
)
@click.option(
    "--dbt-manifest",
    required=True,
    type=click.Path(exists=True),
    help="Path to dbt manifest containing tests to generate MC monitor config for",
)
@click.option(
    "--test-type",
    required=False,
    multiple=True,
    help="Filter for types of dbt tests to convert. Can pass multiple. Defaults to all."
    " Pass SINGULAR to filter to singular tests without a shared type.",
)
@click.option(
    "--label",
    required=False,
    multiple=True,
    help="Label to apply to all generated MC monitors. Can pass multiple.",
)
@click.pass_obj
def generate_from_dbt_tests(ctx, dbt_manifest, output_path, test_type, label):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        dbt_manifest_path=dbt_manifest,
    ).generate_from_dbt_tests(output_path, test_type, label)



"""
Command line tools
"""
import copy
import json
from typing import Set, Optional

import click
from click import Option, Context

import montecarlodata.settings as settings
from montecarlodata.common.common import normalize_gql


class AdvancedOptions(Option):
    """
    Set mutually exclusive options, groups of required options and hidden prompts.
    Errors out on conflict or any missing values and appends text to help.

    Usage example -
        click.option('--foo' cls=MutuallyExclusiveOptions, mutually_exclusive_options=['bar']),
        click.option('--bar', cls=MutuallyExclusiveOptions, mutually_exclusive_options=['foo'])
    """

    def __init__(self, *args, **kwargs):
        help_text = kwargs.get("help", "")
        self._friendly_mutual_args = (
            self._friendly_required_args
        ) = self._friendly_at_least_one_set = None

        self.mutually_exclusive_options = set(
            kwargs.pop("mutually_exclusive_options", [])
        )
        self.required_with_options = set(kwargs.pop("required_with_options", []))
        self.at_least_one_set = set(kwargs.pop("at_least_one_set", []))
        self.prompt_if_requested = kwargs.pop("prompt_if_requested", False)

        self.values_with_required_options = set(
            kwargs.pop("values_with_required_options", [])
        )
        self.required_options_for_values = set(
            kwargs.pop("required_options_for_values", [])
        )

        # Update help for options and create friendly args (i.e. human readable).
        if self.mutually_exclusive_options:
            self._friendly_mutual_args = self._create_friendly_args(
                options=self.mutually_exclusive_options
            )
            help_text = f"{help_text} This option cannot be used with {self._friendly_mutual_args}."
        if self.required_with_options:
            self._friendly_required_args = self._create_friendly_args(
                options=self.required_with_options
            )
            help_text = f"{help_text} This option requires setting {self._friendly_required_args}."
        if self.required_options_for_values:
            self._friendly_required_options_for_values = self._create_friendly_args(
                self.required_options_for_values
            )
            help_text = (
                f"{help_text} This option requires setting {self._friendly_required_options_for_values} "
                f"when it is set to one of these values: {self.values_with_required_options}."
            )
        self._friendly_at_least_one_set = self._create_friendly_args(
            self.at_least_one_set
        )
        kwargs["help"] = help_text
        super(AdvancedOptions, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        """
        Handle any conflicts, missing values or any required hidden prompts.
        """
        cli_friendly_name = normalize_gql(self.name)

        if self.name in opts:
            if self.mutually_exclusive_options.intersection(opts):
                raise click.BadParameter(
                    f"Cannot use '{cli_friendly_name}' with {self._friendly_mutual_args}."
                )
            if not self.required_with_options.issubset(opts):
                raise click.BadParameter(
                    f"Cannot use '{cli_friendly_name}' without {self._friendly_required_args}."
                )
            if opts[
                self.name
            ] in self.values_with_required_options and not self.required_options_for_values.issubset(
                opts
            ):
                raise click.BadParameter(
                    f"Cannot use '{cli_friendly_name}' with value '{opts[self.name]}' without "
                    f"{self._friendly_required_options_for_values}."
                )
            if (
                self.prompt_if_requested
                and opts[self.name] == settings.SHOW_PROMPT_VALUE
            ):
                opts[self.name] = click.prompt(cli_friendly_name, hide_input=True)
            else:
                self.prompt = None

        if self.at_least_one_set and not any(x in opts for x in self.at_least_one_set):
            raise click.BadParameter(
                f"Missing required options from {self._friendly_at_least_one_set}."
            )

        return super(AdvancedOptions, self).handle_parse_result(ctx, opts, args)

    @staticmethod
    def _create_friendly_args(options: Set[str]) -> str:
        """
        Build a more human readable sentence from options.
        """
        friendly_options = list(copy.deepcopy(options))
        last_element = friendly_options.pop() if len(friendly_options) > 1 else None

        friendly_options = ", ".join(
            [f"'{normalize_gql(option)}'" for option in friendly_options]
        )
        if last_element:
            friendly_options = f"{friendly_options}, and '{last_element}'"
        return friendly_options


def add_common_options(options):
    """
    Convenience decorator for shared options (i.e. options that are common across commands)
    """

    def _add_common_options(function):
        for option in options[::-1]:
            function = option(function)
        return function

    return _add_common_options


def validate_json_callback(ctx_, param_, value):
    """
    Convenience callback to help validate (and load) JSON in option strings
    """
    try:
        return json.loads(value)
    except json.decoder.JSONDecodeError as err:
        raise click.BadParameter(f"Malformed JSON - {err}")
    except TypeError:
        return value


def convert_uuid_callback(ctx_, param_, value):
    """
    Convenience callback to convert UUIDs into strings
    """
    if value:
        return str(value)  # str(None) returns 'None'
    return value


def convert_empty_str_callback(ctx_, param_, value):
    """
    Convenience callback to convert an empty (or empty like) string into None
    """
    if bool(value and not value.isspace()):
        return value


def dump_help(cmd, parent: Optional[Context] = None):
    """
    Recursively echos help text for a cmd
    """
    ctx = click.core.Context(cmd, info_name=cmd.name, parent=parent)
    click.echo(cmd.get_help(ctx), err=True)
    for sub in getattr(cmd, "commands", {}).values():
        dump_help(sub, ctx)



import copy
import uuid
from unittest import TestCase
from unittest.mock import Mock, patch

import click
import responses
from box import Box
from pycarlo.common.retries import Backoff
from pycarlo.core import Client, Session

from montecarlodata.collector.fields import ADD_DC_PROMPT_VERBIAGE
from montecarlodata.collector.management import CollectorManagementService
from montecarlodata.common.data import MonolithResponse
from montecarlodata.common.user import UserService
from montecarlodata.queries.collector import ADD_COLLECTOR_RECORD
from montecarlodata.utils import GqlWrapper, AwsClientWrapper
from tests.helpers import capture_function
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_DC_ID = "fb045f19-51c5-4b7b-926b-1defcce05180"
_SAMPLE_TRACE_ID = "09f1fef9-3e58-49cd-8d03-3e8d68491a36"
_SAMPLE_STACK_ARN = "stack_arn"
_SAMPLE_GATEWAY_ID = "gateway_id"
_SAMPLE_TEMPLATE_URI = (
    "https://s3.amazonaws.com/bucket/public/customer_templates/id.json"
)
_SAMPLE_INVALID_LAUNCH_URL = (
    f"https://test.com/cf/home?"
    f"region=us-east-1#/stacks/create/review?stackName=monte-carlo&malformed"
)
_SAMPLE_VALID_LAUNCH_URL = (
    f"https://test.com/cf/home?"
    f"region=us-east-1#/stacks/create/review?"
    f"stackName=monte-carlo&templateURL={_SAMPLE_TEMPLATE_URI}"
)
_SAMPLE_COLLECTOR_RESPONSE = {
    "generateCollectorTemplate": {
        "dc": {
            "uuid": _SAMPLE_DC_ID,
            "templateLaunchUrl": _SAMPLE_VALID_LAUNCH_URL,
            "stackArn": _SAMPLE_STACK_ARN,
            "active": True,
            "apiGatewayId": _SAMPLE_GATEWAY_ID,
            "templateVariant": "dionysus",
        }
    }
}
_SAMPLE_EXTRACTED_COLLECTOR_RESPONSE = _SAMPLE_COLLECTOR_RESPONSE[
    "generateCollectorTemplate"
]
_SAMPLE_MONOLITH_RESPONSE = MonolithResponse(data=_SAMPLE_EXTRACTED_COLLECTOR_RESPONSE)
_SAMPLE_COLLECTORS_RESPONSE = [
    {
        "uuid": "1234",
        "stackArn": "aws-loki",
        "active": True,
        "customerAwsAccountId": "test",
        "templateProvider": "cloudformation",
        "templateVariant": "dionysus",
        "templateVersion": "42",
        "codeVersion": "2042",
        "lastUpdated": None,
    },
    {
        "uuid": "5678",
        "stackArn": "aws-thor",
        "active": True,
        "templateProvider": "terraform",
        "templateVariant": "foo",
        "templateVersion": "42",
        "codeVersion": "2042",
        "lastUpdated": None,
    },
    {
        "uuid": "0912",
        "stackArn": "aws-odin",
        "active": False,
        "templateProvider": None,
        "templateVariant": None,
        "templateVersion": "42",
        "codeVersion": "2042",
        "lastUpdated": None,
    },
]


class MockBackoff(Backoff):
    def __init__(self, count: int):
        super(MockBackoff, self).__init__(start=0, maximum=0)
        self.count = count

    def backoff(self, attempt: int) -> float:
        return 0

    def delays(self):
        for i in range(self.count):
            yield 0


class CollectorManagementTest(TestCase):
    def setUp(self) -> None:
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)
        self._user_service_mock = Mock(autospec=UserService)
        self._mc_client = Client(
            session=Session(
                endpoint=_SAMPLE_CONFIG.mcd_api_endpoint,
                mcd_id=_SAMPLE_CONFIG.mcd_id,
                mcd_token=_SAMPLE_CONFIG.mcd_token,
            )
        )

        self._service = CollectorManagementService(
            _SAMPLE_CONFIG,
            mc_client=self._mc_client,
            retry_policy=MockBackoff(1),
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )

    def test_get_template_with_response(self):
        self._request_wrapper_mock.make_request_v2.return_value = None
        with self.assertRaises(click.exceptions.Abort):
            self._service.echo_template()

    def test_get_template_with_invalid_link(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"dc": {"templateLaunchUrl": _SAMPLE_INVALID_LAUNCH_URL}}
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service.echo_template()

    def test_get_template_with_valid_link(self):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={"dc": {"templateLaunchUrl": _SAMPLE_VALID_LAUNCH_URL}}
        )

        std_out = capture_function(function=self._service.echo_template).std_out
        self.assertEqual(std_out.getvalue().strip(), _SAMPLE_TEMPLATE_URI)

    def test_upgrade_with_inactive_collector(self):
        sample_collector_response = copy.deepcopy(_SAMPLE_COLLECTOR_RESPONSE)
        sample_collector_response["generateCollectorTemplate"]["dc"]["active"] = False

        self._request_wrapper_mock.make_request.return_value = sample_collector_response
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

    def test_upgrade_with_missing_fields(self):
        self._request_wrapper_mock.make_request.return_value = {
            "generateCollectorTemplate": {"dc": {}}
        }
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

    @patch("montecarlodata.collector.management.uuid")
    @responses.activate
    def test_upgrade(self, mock_uuid):
        mock_uuid.uuid4.return_value = _SAMPLE_TRACE_ID
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = True
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._user_service_mock.get_collector.return_value = Box(
            {"uuid": _SAMPLE_DC_ID}
        )
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "pingDataCollector": {
                        "dcId": _SAMPLE_DC_ID,
                        "traceId": _SAMPLE_TRACE_ID,
                    },
                }
            },
        )

        self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )
        mock_uuid.uuid4.assert_called_once()

    @patch("montecarlodata.collector.management.uuid")
    @responses.activate
    def test_upgrade_ping_fail(self, mock_uuid):
        mock_uuid.uuid4.return_value = _SAMPLE_TRACE_ID
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = True
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={"errors": [{"message": "Data collector not found"}]},
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )
        mock_uuid.uuid4.assert_called_once()

    @patch("montecarlodata.collector.management.uuid")
    @responses.activate
    def test_upgrade_ping_timeout(self, mock_uuid):
        mock_uuid.uuid4.return_value = _SAMPLE_TRACE_ID
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = True
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        responses.post(_SAMPLE_CONFIG.mcd_api_endpoint, status=504)

        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )
        mock_uuid.uuid4.assert_called_once()

    @responses.activate
    def test_upgrade_ping_mismatched_trace_id(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = True
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "pingDataCollector": {
                        "dcId": _SAMPLE_DC_ID,
                        "traceId": str(uuid.uuid4()),
                    },
                }
            },
        )

        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )

    @patch("montecarlodata.collector.management.uuid")
    @responses.activate
    def test_upgrade_to_janus(self, mock_uuid):
        mock_uuid.uuid4.return_value = _SAMPLE_TRACE_ID
        self._aws_wrapper_mock.get_stack_parameters.return_value = [
            {"ParameterKey": "CreateCrossAccountRole", "ParameterValue": "True"},
            {"ParameterKey": "CreateEventInfra", "ParameterValue": "True"},
            {"ParameterKey": "EnableGatewayAccessLogging", "ParameterValue": "False"},
            {"ParameterKey": "ExistingS3Vpce", "ParameterValue": "True"},
            {"ParameterKey": "ExistingSubnetIds", "ParameterValue": "True"},
            {"ParameterKey": "ExistingVpcId", "ParameterValue": "True"},
            {"ParameterKey": "subnetCIDRs", "ParameterValue": "2222"},
            {"ParameterKey": "vpcCIDR", "ParameterValue": "1111"},
            {
                "ParameterKey": "workerLambdaConcurrentExecutions",
                "ParameterValue": "12",
            },
        ]
        self._aws_wrapper_mock.upgrade_stack.return_value = True
        self._user_service_mock.get_collector.return_value = Box(
            {"uuid": _SAMPLE_DC_ID}
        )
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "pingDataCollector": {
                        "dcId": _SAMPLE_DC_ID,
                        "traceId": _SAMPLE_TRACE_ID,
                    },
                }
            },
        )

        self._service.upgrade_template(
            update_infra=True,
            new_params={"EnableRemoteUpdates": "True", "CreateEventInfra": "False"},
        )

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=[
                # updated parameter
                {
                    "ParameterKey": "CreateEventInfra",
                    "ParameterValue": "False",
                    "UsePreviousValue": False,
                },
                # preserved parameters
                {"ParameterKey": "ExistingS3Vpce", "UsePreviousValue": True},
                {"ParameterKey": "ExistingSubnetIds", "UsePreviousValue": True},
                {"ParameterKey": "ExistingVpcId", "UsePreviousValue": True},
                # migrated parameters
                {"ParameterKey": "SubnetCIDRs", "ParameterValue": "2222"},
                {"ParameterKey": "VpcCIDR", "ParameterValue": "1111"},
                {
                    "ParameterKey": "WorkerLambdaConcurrentExecutions",
                    "ParameterValue": "12",
                },
                # new parameter
                {"ParameterKey": "EnableRemoteUpdates", "ParameterValue": "True"},
            ],
        )
        self._aws_wrapper_mock.deploy_gateway.assert_called_once_with(
            gateway_id=_SAMPLE_GATEWAY_ID
        )

    def test_failed_stack_upgrade(self):
        self._aws_wrapper_mock.get_stack_parameters.return_value = None
        self._aws_wrapper_mock.upgrade_stack.return_value = False

        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service.upgrade_template()

        self._aws_wrapper_mock.get_stack_parameters.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN
        )
        self._aws_wrapper_mock.upgrade_stack.assert_called_once_with(
            stack_id=_SAMPLE_STACK_ARN,
            template_link=_SAMPLE_TEMPLATE_URI,
            parameters=None,
        )
        self._aws_wrapper_mock.deploy_gateway.assert_not_called()

    def test_deploy(self):
        stack_name, termination_protection = "foo", False
        collector_response = copy.deepcopy(_SAMPLE_COLLECTOR_RESPONSE)
        collector_response["generateCollectorTemplate"]["dc"]["active"] = False
        monolith_response = MonolithResponse(
            data=collector_response["generateCollectorTemplate"]
        )

        self._aws_wrapper_mock.create_stack.return_value = True

        self._request_wrapper_mock.make_request_v2.return_value = monolith_response
        self._service.deploy_template(
            stack_name=stack_name, termination_protection=termination_protection
        )

        self._aws_wrapper_mock.create_stack.assert_called_once_with(
            stack_name=stack_name,
            template_link=_SAMPLE_TEMPLATE_URI,
            termination_protection=termination_protection,
            parameters=[],
        )

    def test_deploy_with_an_active_collector(self):
        stack_name, termination_protection = "foo", True
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        with self.assertRaises(click.exceptions.Abort):
            self._service.deploy_template(
                stack_name=stack_name, termination_protection=termination_protection
            )

    def test_replace_params(self):
        existing_params, new_params = [
            {"ParameterKey": "foo", "ParameterValue": "6"}
        ], {"foo": "42"}
        parameters = self._service._build_param_list(
            existing_params=existing_params, new_params=new_params
        )
        self.assertEqual(
            parameters,
            [
                {
                    "ParameterKey": "foo",
                    "ParameterValue": "42",
                    "UsePreviousValue": False,
                }
            ],
        )

    def test_only_existing_params(self):
        existing_params = [{"ParameterKey": "foo", "ParameterValue": "42"}]
        parameters = self._service._build_param_list(
            existing_params=existing_params, new_params=None
        )
        self.assertEqual(
            parameters, [{"ParameterKey": "foo", "UsePreviousValue": True}]
        )

    def test_completely_new_params(self):
        parameters = self._service._build_param_list(
            existing_params=[], new_params={"foo": "42"}
        )
        self.assertEqual(parameters, [{"ParameterKey": "foo", "ParameterValue": "42"}])

    @patch("montecarlodata.collector.management.click")
    def test_get_launch_link(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._service.launch_quick_create_link(dry=False)

        click_mock.launch.assert_called_once_with(_SAMPLE_VALID_LAUNCH_URL)
        click_mock.echo.assert_not_called()

    @patch("montecarlodata.collector.management.click")
    def test_get_launch_link_dry_run(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = (
            _SAMPLE_MONOLITH_RESPONSE
        )
        self._service.launch_quick_create_link(dry=True)

        click_mock.echo.assert_called_once_with(_SAMPLE_VALID_LAUNCH_URL)
        click_mock.launch.assert_not_called()

    def test_set_region(self):
        region = "loki"
        service = CollectorManagementService(
            _SAMPLE_CONFIG,
            mc_client=self._mc_client,
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            aws_region_override=region,
        )
        self.assertEqual(service._collector_region, region)
        self.assertEqual(self._service._collector_region, "us-east-1")
        self.assertEqual(self._service._collector_region, _SAMPLE_CONFIG.aws_region)

    @patch("montecarlodata.collector.management.click")
    def test_echo_collectors(self, click_mock):
        self._user_service_mock.collectors = _SAMPLE_COLLECTORS_RESPONSE

        self._service.echo_collectors(table_format="plain")
        click_mock.echo.assert_called_once_with(
            "AWS Stack ARN      ID    Version  Template                    Last updated    Active\n"
            "aws-loki         1234       2042  cloudformation:dionysus:42  -               True\n"
            "aws-thor         5678       2042  terraform:foo:42            -               True\n"
            "aws-odin         0912       2042  cloudformation:July-2019    -               False"
        )

    @patch("montecarlodata.collector.management.click")
    def test_echo_collectors_with_only_active(self, click_mock):
        self._user_service_mock.collectors = _SAMPLE_COLLECTORS_RESPONSE

        self._service.echo_collectors(table_format="plain", active_only=True)
        click_mock.echo.assert_called_once_with(
            "AWS Stack ARN      ID    Version  Template                    Last updated    Active\n"
            "aws-loki         1234       2042  cloudformation:dionysus:42  -               True\n"
            "aws-thor         5678       2042  terraform:foo:42            -               True"
        )

    @patch.object(CollectorManagementService, "launch_quick_create_link")
    @patch("montecarlodata.collector.management.prompt_connection")
    @patch("montecarlodata.collector.management.click")
    def test_add_collector(
        self, click_mock, prompt_connection_mock, launch_quick_create_link_mock
    ):
        mocked_region = "us-east-1"
        click_mock.prompt.return_value = mocked_region
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"dc": {"uuid": "1234"}})
        )

        self._service.add_collector()

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_COLLECTOR_RECORD, operation="createCollectorRecord"
        )
        click_mock.echo.assert_called_once_with(
            "Created collector record with ID '1234'"
        )
        prompt_connection_mock.assert_called_once_with(message=ADD_DC_PROMPT_VERBIAGE)
        launch_quick_create_link_mock.assert_called_once_with(
            dry=False, dc_id="1234", collection_region=mocked_region
        )

    @patch("montecarlodata.collector.management.click")
    def test_add_collector_no_prompt(self, click_mock):
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data=Box({"dc": {"uuid": "1234"}})
        )

        self._service.add_collector(no_prompt=True)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=ADD_COLLECTOR_RECORD,
            operation="createCollectorRecord",
        )
        click_mock.echo.assert_called_once_with(
            "Created collector record with ID '1234'"
        )
